"""_import_vdb_all_grids.py: Module to import a vdb into resqml format."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.vdb as vdb
import resqpy.rq_import as rqi


def import_vdb_all_grids(
    resqml_file_root,  # output path and file name without .epc or .h5 extension
    extent_ijk = None,  # 3 element numpy vector applicable to ROOT
    vdb_file = None,
    vdb_case = None,  # if None, first case in vdb is used (usually a vdb only holds one case)
    corp_xy_units = 'm',
    corp_z_units = 'm',
    corp_z_inc_down = True,
    ijk_handedness = 'right',
    geometry_defined_everywhere = True,
    treat_as_nan = None,
    resqml_xy_units = 'm',
    resqml_z_units = 'm',
    resqml_z_inc_down = True,
    shift_to_local = False,
    local_origin_place = 'centre',  # 'centre' or 'minimum'
    max_z_void = 0.1,  # vertical gaps greater than this will introduce k gaps intp resqml grid
    split_pillars = True,
    split_tolerance = 0.01,  # applies to each of x, y, z differences
    vdb_static_properties = True,
    # if True, static vdb properties are imported (only relevant if vdb_file is not None)
    vdb_recurrent_properties = False,
    decoarsen = True,
    timestep_selection = 'all',
    # 'first', 'last', 'first and last', 'all', or list of ints being reporting timestep numbers
    create_property_set = False):
    """Creates a RESQML dataset containing grids and grid properties, including LGRs, for a single realisation.

    Arguments:
        resqml_file_root (str): output path and file name without .epc or .h5 extension
        extent_ijk (triple float, optional): ijk extents (fortran ordering)
        vdb_file (str, optional): vdb input file, either this or corp_file should be not None. Required if importing from a vdb
        vdb_case (str, optional): required if the vdb contains more than one case. If None, first case in vdb is used
        corp_xy_units (str, default 'm'): xy length units
        corp_z_units (str, default 'm'): z length units
        corp_z_inc_down (bool, default True): if True z values increase with depth
        ijk_handedness (str, default 'right'): 'right' or 'left'
        geometry_defined_everywhere (bool, default True): if False then inactive cells are marked as not having geometry
        treat_as_nan (float, default None): if a value is provided corner points with this value will be assigned nan
        resqml_xy_units (str, default 'm'): output xy units for resqml file
        resqml_z_units (str, default 'm'): output z units for resqml file
        resqml_z_inc_down (bool, default True): if True z values increase with depth for output resqml file
        shift_to_local (bool, default False): if True then a local origin will be used in the CRS
        local_origin_place (str, default 'centre'): 'centre' or 'minimum'. If 'centre' the local origin is placed at the centre of the grid; ignored if shift_to_local is False
        max_z_void (float, default 0.1): maximum z gap between vertically neighbouring corner points. Vertical gaps greater than this will introduce k gaps into resqml grid. Units are corp z units
        split_pillars (bool, default True): if False an unfaulted grid will be generated
        split_tolerance (float, default 0.01): maximum distance between neighbouring corner points before a pillar is considered 'split'. Applies to each of x, y, z differences
        vdb_static_properties (bool, default True): if True, static vdb properties are imported (only relevant if vdb_file is not None)
        vdb_recurrent_properties (bool, default False): # if True, recurrent vdb properties are imported (only relevant if vdb_file is not None)
        decoarsen (bool, default True): where ICOARSE is present, redistribute data to uncoarse cells
        timestep_selection (str, default 'all): 'first', 'last', 'first and last', 'all', or list of ints being reporting timestep numbers. Ignored if vdb_recurrent_properties is False
        create_property_set (bool, default False): if True a resqml PropertySet is created
    """

    vdbase = vdb.VDB(vdb_file)
    case_list = vdbase.cases()
    assert len(case_list) > 0, 'no cases found in vdb'
    if vdb_case is None:
        vdb_case = case_list[0]
    else:
        assert vdb_case in case_list, 'case ' + vdb_case + ' not found in vdb: ' + vdb_file
        vdbase.set_use_case(vdb_case)
    grid_list = vdbase.list_of_grids()
    index = 0
    for grid_name in grid_list:
        if grid_name.upper().startswith('SMALLGRIDS'):
            log.warning('vdb import skipping small grids')
            continue
        log.debug('importing vdb data for grid ' + str(grid_name))
        rqi.import_nexus(
            resqml_file_root,
            extent_ijk = extent_ijk if grid_name == 'ROOT' else None,  # 3 element numpy vector applicable to ROOT
            vdb_file = vdb_file,
            vdb_case = vdb_case,  # if None, first case in vdb is used (usually a vdb only holds one case)
            corp_xy_units = corp_xy_units,
            corp_z_units = corp_z_units,
            corp_z_inc_down = corp_z_inc_down,
            ijk_handedness = ijk_handedness,
            geometry_defined_everywhere = geometry_defined_everywhere,
            treat_as_nan = treat_as_nan,
            resqml_xy_units = resqml_xy_units,
            resqml_z_units = resqml_z_units,
            resqml_z_inc_down = resqml_z_inc_down,
            shift_to_local = shift_to_local,
            local_origin_place = local_origin_place,  # 'centre' or 'minimum'
            max_z_void = max_z_void,  # vertical gaps greater than this will introduce k gaps intp resqml grid
            split_pillars = split_pillars,  # NB: some LGRs may be unsplit even if ROOT is split
            split_tolerance = split_tolerance,  # applies to each of x, y, z differences
            vdb_static_properties = vdb_static_properties,  # if True, static vdb properties are imported
            vdb_recurrent_properties = vdb_recurrent_properties,
            decoarsen = decoarsen,
            timestep_selection = timestep_selection,
            create_property_set = create_property_set,
            grid_title = grid_name,
            mode = 'w' if index == 0 else 'a')
        index += 1



"""Classes for storing and retrieving dataframes as RESQML objects.

Note that this module uses the obj_Grid2dRepresentation class in a way that was not envisaged
when the RESQML standard was defined; software that does not use resqpy is unlikely to be
able to do much with data stored in this way
"""

import logging

log = logging.getLogger(__name__)

import warnings
import numpy as np
import pandas as pd

import resqpy.crs as rqc
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
import resqpy.surface as rqs
import resqpy.time_series as rqts

# todo: add support for building an ensemble of dataframes using the same mesh support


class DataFrame:
    """Class for storing and retrieving a pandas dataframe of numerical data as a RESQML property.

    notes:
       actual values are stored either as z values in a Mesh (Grid2d) object, or as a property on
       such a mesh when multiple raalizations are in use; a regular Mesh object is created to act
       as a supporting representation; columns are mapped onto I and rows onto J; if a property is
       used then the indexable elements are 'nodes'; column titles are stored in a related StringLookup
       object, indexed by column number; column units are optionally treated in the same way (uom for
       the property is generally set to Euc); all values are stored as floats; use the derived TimeTable
       class if rows relate to steps in a TimeSeries; use the derived RelPerm class if the rows relate to
       relative permeability data
    """

    def __init__(self,
                 model,
                 uuid = None,
                 df = None,
                 uom_list = None,
                 realization = None,
                 title = 'dataframe',
                 column_lookup_uuid = None,
                 uom_lookup_uuid = None,
                 extra_metadata = None):
        """Create a new Dataframe object from either a previously stored property or a pandas dataframe.

        arguments:
           model (model.Model): the model to which the new Dataframe will be attached
           uuid (uuid.UUID, optional): the uuid of an existing Grid2dRepresentation
              object acting as support for a dataframe property (or holding the dataframe as z values)
           df (pandas.DataFrame, optional): a dataframe from which the new Dataframe is to be created;
              if both uuid and df are supplied, realization must not be None and a new
              realization property will be created
           uom_list (list of str, optional): a list holding the units of measure for each
              column; if present, length of list must match number of columns in df; ignored if
              uuid is not None
           realization (int, optional): if present, the realization number of the RESQML property
              holding the dataframe
           title (str, default 'dataframe'): used as the citation title for the Mesh (and property);
              ignored if uuid is not None
           column_lookup_uuid (uuid, optional): if present, the uuid of a string lookup table holding
              the column names; if present, the contents and order of the table must match the columns
              in the dataframe; if absent, a new lookup table will be created
           uom_lookup_uuid (uuid, optional): if present, the uuid of a string lookup table holding
              the units of measure for each column; if None and uom_list is present, a new table
              will be created; if both uom_list and uom_lookup_uuid are present, their contents
              must match
           extra_metadata (dict, optional): if present, a dictionary of extra metadata items, str: str;
              ignored if uuid is not None

        returns:
           a newly created Dataframe object

        notes:
           when initialising from an existing RESQML object, the supporting mesh and its property should
           have been originally created using this class; when working with ensembles, each object of this
           class will only handle the data for one realization, though they may share a common support;
           if both a uuid and a df are provided, a realization number must also be given and the dataframe
           is used to create a new realization similar to that identified by the uuid
        """

        assert uuid is not None or df is not None
        assert uuid is None or df is None or realization is not None

        self.model = model
        self.df = None
        self.n_rows = self.n_cols = 0
        self.uom_list = None
        self.realization = realization
        self.title = title
        self.mesh = None  # only generated when needed for write_hdf5(), create_xml()
        self.pc = None  # property collection; only generated when needed for write_hdf5(), create_xml()
        self.column_lookup_uuid = column_lookup_uuid
        self.column_lookup = None  # string lookup table mapping column index (0 based) to column name
        self.uom_lookup_uuid = uom_lookup_uuid
        self.uom_lookup = None  # string lookup table mapping column index (0 based) to uom
        self.extra_metadata = extra_metadata

        if uuid is not None:
            support_root = model.root_for_uuid(uuid)
            assert rqet.node_type(support_root) == 'obj_Grid2dRepresentation'
            self.mesh = rqs.Mesh(self.model, uuid = uuid)
            self.extra_metadata = self.mesh.extra_metadata
            assert 'dataframe' in self.extra_metadata and self.extra_metadata['dataframe'] == 'true'
            self.title = self.mesh.title
            self.n_rows, self.n_cols = self.mesh.nj, self.mesh.ni
            cl_uuid = self.model.uuid(obj_type = 'StringTableLookup', related_uuid = uuid, title = 'dataframe columns')
            assert cl_uuid is not None, 'column name lookup table not found for dataframe'
            self.column_lookup = rqp.StringLookup(self.model, uuid = cl_uuid)
            self.column_lookup_uuid = self.column_lookup.uuid
            assert self.column_lookup.length() == self.n_cols
            ul_uuid = self.model.uuid(obj_type = 'StringTableLookup', related_uuid = uuid, title = 'dataframe units')
            if ul_uuid is not None:
                self.uom_lookup = rqp.StringLookup(self.model, uuid = ul_uuid)
                self.uom_lookup_uuid = self.uom_lookup.uuid
                self.uom_list = self.uom_lookup.get_list()
            da = self.mesh.full_array_ref()[..., 2]  # dataframe data as 2D numpy array, defaulting to z values in mesh
            existing_pc = rqp.PropertyCollection(support = self.mesh)
            existing_count = 0 if existing_pc is None else existing_pc.number_of_parts()
            if df is None:  # existing dara, either in mesh or property
                if existing_count > 0:  # use property data instead of z values
                    if existing_count == 1:
                        if self.realization is not None:
                            assert existing_pc.realization_for_part(existing_pc.singleton()) == self.realization
                    else:
                        assert self.realization is not None, 'no realization specified when accessing ensemble dataframe'
                    da = existing_pc.single_array_ref(realization = self.realization)
                    assert da is not None and da.ndim == 2 and da.shape == (self.n_rows, self.n_cols)
                else:
                    assert realization is None
                self.df = pd.DataFrame(da, columns = self.column_lookup.get_list())
            else:  # both uuid and df supplied: add a new realisation
                if existing_count > 0:
                    assert existing_pc.singleton(
                        realization = self.realization) is None, 'dataframe realization already exists'
                self.df = df.copy()
                assert len(self.df) == self.n_rows
                assert len(self.df.columns) == self.n_rows
        else:
            assert df is not None, 'no dataframe provided when instantiating DataFrame object'
            self.df = df.copy()
            # todo: check data type of columns – restrict to numerical data
            self.n_rows = len(self.df)
            self.n_cols = len(self.df.columns)
            if column_lookup_uuid is not None:
                self.column_lookup = rqp.StringLookup(self.model, uuid = column_lookup_uuid)
                assert self.column_lookup is not None
                assert self.column_lookup.length() == self.n_cols
                assert all(self.df.columns == self.column_lookup.get_list())  # exact match of column names required!
            if uom_lookup_uuid is not None:
                self.uom_lookup = rqp.StringLookup(self.model, uuid = uom_lookup_uuid)
                assert self.uom_lookup is not None
            if uom_list is not None:
                assert len(uom_list) == self.n_cols
                self.uom_list = uom_list.copy()
                if self.uom_lookup is not None:
                    assert self.uom_list == self.uom_lookup.get_list()
            elif self.uom_lookup is not None:
                self.uom_list = self.uom_lookup.get_list()

    def dataframe(self):
        """Returns the Dataframe as a pandas DataFrame."""

        return self.df

    def column_uom(self, col_index):
        """Returns units of measure for the specified column, or Euc if no units present."""

        if self.units_table is None:
            return 'Euc'
        assert 0 <= col_index < self.n_cols, 'column index out of range'
        return self.units_table.get_string(col_index)

    def write_hdf5_and_create_xml(self):
        """Write dataframe data to hdf5 file and create xml for RESQML objects to represent dataframe."""

        self._set_mesh_from_df()  # writes hdf5 data and creates xml for mesh (and property)

        if self.column_lookup is None:
            self.column_lookup = rqp.StringLookup(self.model,
                                                  int_to_str_dict = dict(enumerate(self.df.columns)),
                                                  title = 'dataframe columns')
            self.column_lookup_uuid = self.column_lookup.uuid
            sl_node = self.column_lookup.create_xml()
        else:
            sl_node = self.column_lookup.root
        if sl_node is not None:
            self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', sl_node, 'sourceObject')

        if self.uom_list and self.uom_lookup is None:
            self.uom_lookup = rqp.StringLookup(self.model,
                                               int_to_str_dict = dict(enumerate(self.uom_list)),
                                               title = 'dataframe units')
            self.uom_lookup_uuid = self.uom_lookup.uuid
            ul_node = self.uom_lookup.create_xml()
        elif self.uom_lookup is not None:
            ul_node = self.uom_lookup.root
        else:
            ul_node = None
        if ul_node is not None:
            self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', ul_node, 'sourceObject')

    def _set_mesh_from_df(self):
        """Creates Mesh object; called before writing to hdf5 or creating xml."""
        # note: actual data is stored in related Property if realization number is present, directly in Mesh otherwise

        assert self.n_rows == len(self.df)
        assert self.n_cols == len(self.df.columns)

        if self.mesh is None:
            origin = (0.0, 0.0, 0.0)
            dxyz_dij = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
            crs_uuids = self.model.uuids(obj_type = 'LocalDepth3dCrs')
            if len(crs_uuids) == 0:
                crs = rqc.Crs(self.model)
                crs.create_xml()
                crs_uuid = crs.uuid
            else:  # use any available crs
                crs_uuid = crs_uuids[0]
            if self.realization is None:
                self.mesh = rqs.Mesh(self.model,
                                     mesh_flavour = 'reg&z',
                                     ni = self.n_cols,
                                     nj = self.n_rows,
                                     dxyz_dij = dxyz_dij,
                                     origin = origin,
                                     z_values = np.array(self.df),
                                     crs_uuid = crs_uuid)
            else:
                self.mesh = rqs.Mesh(self.model,
                                     mesh_flavour = 'regular',
                                     ni = self.n_cols,
                                     nj = self.n_rows,
                                     dxyz_dij = dxyz_dij,
                                     origin = origin,
                                     crs_uuid = crs_uuid)
            self.mesh.write_hdf5()
            mesh_root = self.mesh.create_xml(title = self.title)
            rqet.create_metadata_xml(mesh_root, {'dataframe': 'true'})
            if self.realization is not None:
                self.pc = rqp.PropertyCollection()
                self.pc.set_support(support = self.mesh)
                dataframe_pk_uuid = self.model.uuid(obj_type = 'PropertyKind', title = 'dataframe')
                if dataframe_pk_uuid is None:
                    dataframe_pk = rqp.PropertyKind(self.model, title = 'dataframe', example_uom = 'Euc')
                    dataframe_pk.create_xml()
                    dataframe_pk_uuid = dataframe_pk.uuid
                self.pc.add_cached_array_to_imported_list(np.array(self.df),
                                                          'dataframe',
                                                          self.title,
                                                          uom = 'Euc',
                                                          property_kind = 'dataframe',
                                                          local_property_kind_uuid = dataframe_pk_uuid,
                                                          realization = self.realization,
                                                          indexable_element = 'nodes')
                self.pc.write_hdf5_for_imported_list()
                self.pc.create_xml_for_imported_list_and_add_parts_to_model()


class TimeTable(DataFrame):
    """Class for storing and retrieving a pandas dataframe where rows relate to steps in a time series.

    note:
       inherits from DataFrame class
    """

    def __init__(self,
                 model,
                 uuid = None,
                 df = None,
                 uom_list = None,
                 realization = None,
                 time_series = None,
                 title = 'timetable',
                 column_lookup_uuid = None,
                 uom_lookup_uuid = None):
        """Create a new TimeTable object from either a previously stored property or a pandas dataframe.

        arguments:
           time_series (resqpy.time_series.TimeSeries): the time series which rows in the dataframe relate to;
              required if initialising from a dataframe, ignored otherwise

        note:
           see DataFrame class docstring for details of other arguments
        """

        # todo: add option to set up time series from a column in the dataframe?

        assert uuid is not None or (df is not None and time_series is not None)

        super().__init__(model,
                         uuid = uuid,
                         df = df,
                         uom_list = uom_list,
                         realization = realization,
                         title = title,
                         column_lookup_uuid = column_lookup_uuid,
                         uom_lookup_uuid = uom_lookup_uuid)
        if uuid is not None:
            ts_uuid = self.model.uuid(obj_type = 'TimeSeries', related_uuid = self.mesh.uuid)
            assert ts_uuid is not None, 'no time series related to mesh holding dataframe'
            self.ts = rqts.TimeSeries(self.model, uuid = ts_uuid)
        else:
            assert time_series is not None
            assert time_series.number_of_timestamps() == self.n_rows
            self.ts = time_series

    def time_series(self):
        """Returns the TimeSeries object in use by the time table."""

        return self.ts

    def write_hdf5_and_create_xml(self):
        """Write time table data to hdf5 file and create xml for RESQML objects to represent dataframe."""

        super().write_hdf5_and_create_xml()
        # note: time series xml must be created before calling this method
        self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', self.ts.root, 'sourceObject')


def dataframe_parts_in_model(model, timetables = None, title = None, related_uuid = None):
    """Returns list of part names within model that are representing DataFrame support objects.

    arguments:
       model (model.Model): the model to be inspected for dataframes
       timetables (boolean or None): if True, only TimeTable dataframe parts will be included; if False
          only DataFrame parts that are not representing TimeTable objects will be included; if None,
          both parts for both types of dataframe will be included
       title (str, optional): if present, only parts with a citation title exactly matching will be
          included
       related_uuid (uuid, optional): if present, only parts relating to this uuid are included

    returns:
       list of str, each element in the list is a part name, within model, which is representing the
       support for a DataFrame object
    """

    df_parts_list = model.parts(obj_type = 'Grid2dRepresentation',
                                title = title,
                                extra = {'dataframe': 'true'},
                                related_uuid = related_uuid)

    if timetables is not None:
        filtered_list = []
        for df_part in df_parts_list:
            is_tt = (model.part(obj_type = 'TimeSeries', related_uuid = model.uuid_for_part(df_part)) is not None)
            if timetables == is_tt:
                filtered_list.append(df_part)
        df_parts_list = filtered_list

    return df_parts_list


def timetable_parts_in_model(model, title = None, related_uuid = None):
    """Returns list of part names within model that are representing TimeTable dataframe support objects.

    arguments:
       model (model.Model): the model to be inspected for dataframes
       title (str, optional): if present, only parts with a citation title exactly matching will be
          included
       related_uuid (uuid, optional): if present, only parts relating to this uuid are included

    returns:
       list of str, each element in the list is a part name, within model, which is representing the support
       for a TimeTable object
    """

    return dataframe_parts_in_model(model, timetables = True, title = title, related_uuid = related_uuid)


def dataframe_for_title(model, title, realization = None):
    """Returns a DataFrame object loaded from model, with given title (optionally for given realization)."""

    df_parts = dataframe_parts_in_model(model, title = title)
    if df_parts is None or len(df_parts) == 0:
        return None
    assert len(df_parts) == 1
    return DataFrame(model, uuid = model.uuid_for_part(df_parts[0]), realization = realization)


def timetable_for_title(model, title, realization = None):
    """Returns a TimeTable object loaded from model, with given title (optionally for given realization)."""

    tt_parts = timetable_parts_in_model(model, title = title)
    if tt_parts is None or len(tt_parts) == 0:
        return None
    assert len(tt_parts) == 1
    return TimeTable(model, uuid = model.uuid_for_part(tt_parts[0]), realization = realization)



"""Weights and measures valid units and unit conversion functions."""

__all__ = [
    'UOM_ALIASES', 'UOM_ALIAS_MAP', 'CASE_INSENSITIVE_UOMS', 'rq_uom', 'convert', 'valid_uoms', 'valid_quantities',
    'valid_property_kinds', 'nexus_uom_for_quantity', 'rq_uom_list', 'rq_length_unit', 'rq_time_unit', 'convert_times',
    'convert_lengths', 'convert_pressures', 'convert_volumes', 'convert_flow_rates', 'convert_transmissibilities',
    'get_conversion_factors'
]

from .weights_and_measures import (UOM_ALIASES, UOM_ALIAS_MAP, CASE_INSENSITIVE_UOMS, rq_uom, convert, valid_uoms,
                                   valid_quantities, valid_property_kinds, rq_uom_list, rq_length_unit, rq_time_unit,
                                   convert_times, convert_lengths, convert_pressures, convert_volumes,
                                   convert_flow_rates, convert_transmissibilities, get_conversion_factors)
from .nexus_units import nexus_uom_for_quantity

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""time.py: A very thin wrapper around python datetime functionality, to meet resqml standard."""

import datetime as dt


def now(use_utc = False):
    """Returns an iso format string representation of the current time, to the nearest second.

    argument:
       use_utc (boolean, default False): if True, the current UTC time is used, otherwise local time

    returns:
       string of form YYYY-MM-DDThh:mm:ssZ representing the current time in iso format

    note:
       this is the format used by the resqml standard for representing date-times
    """

    if use_utc:
        stamp = dt.datetime.utcnow()  # NB: naive use of UTC time
    else:
        stamp = dt.datetime.now()
    return str(stamp.isoformat(sep = 'T', timespec = 'seconds')) + 'Z'



"""Function for creating a time series from a Nexus summary file."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import datetime as dt
import os

import resqpy.time_series
import resqpy.time_series._time_series as rqts
import resqpy.weights_and_measures as wam


def time_series_from_nexus_summary(summary_file, parent_model = None, start_date = None):
    """Create a TimeSeries object based on time steps reported in a Nexus summary file (.sum).

    arguments:
        summary_file (str): path of Nexus summary file
        parent_model (Model): the model to which the new time series will be attached
        start_date (str, optional): if the summary file does not contain dates, this will be
            used as the start date; required format is 'YYYY-MM-DD'

    returns:
        newly created TimeSeries

    note:
        this function does not create the xml for the new TimeSeries, nor add it as a part to the parent model
    """

    if not summary_file:
        log.warning('no summary file specified')
        return None

    if not os.path.isfile(summary_file):
        log.error(f'summary file not found: {summary_file}')
        return None

    try:

        summary_entries = _open_file(summary_file)

        time_series = _process_summary_entries(summary_entries, parent_model, start_date = start_date)

    except Exception:
        log.exception(f'failed to create TimeSeries object from summary file: {summary_file}')
        return None

    return time_series


def _process_summary_entries(summary_entries, parent_model = None, start_date = None):
    """Create a TimeSeries object based on time steps reported in a Nexus summary file (.sum)."""
    if start_date is not None:
        assert len(start_date) == 10 and start_date[4] == '-' and start_date[7] == '-',  \
            f'start date {start_date} specified for summary data is not in format YYYY-MM-DD'
        start_date = dt.date(int(start_date[:4]), int(start_date[5:7]), int(start_date[8:]))
    if len(summary_entries) == 0:
        log.warning('no entries read from Nexus summary file')
        return None  # no entries extracted from summary file, could raise error?
    summary_entries.sort()
    # sort out the start date
    tz_date = summary_entries[0][2]
    if tz_date is None:
        if start_date is None:
            log.warning('summary file does not include dates and no start date specified; using arbitrary date')
            tz_date = dt.date(2020, 1, 1)
            start_date = tz_date
        else:
            tz_date = start_date
    else:
        if summary_entries[0][0] > 0:
            delta = dt.timedelta(days = summary_entries[0][1])
            tz_date = tz_date - delta
            # todo: round to the nearest previous midnight?
        if start_date is not None and tz_date != start_date:
            log.warning(f'ignoring specified start date {start_date} in favour of summary file date {tz_date}')
            start_date = tz_date
    if summary_entries[0][0] == 0:  # first entry is for time step zero, handled as first timestamp
        summary_entries.pop(0)
    # intiialise the time series with just the start date
    time_series = rqts.TimeSeries(parent_model = parent_model, first_timestamp = tz_date.isoformat() + 'T00:00:00Z')
    last_timestep = 0
    last_cumulative_time = 0.0
    cumulative_time_error = False
    for entry in summary_entries:
        if entry[0] <= last_timestep:
            log.warning(f'ignoring out of sequence time step in summary file: {entry[0]}')
            continue
        while entry[0] > last_timestep + 1:  # add filler steps
            time_series.extend_by_days(0)
            last_timestep += 1
        if entry[1] < last_cumulative_time:  # should never happen
            time_series.extend_by_days(0)
            cumulative_time_error = True
        else:
            time_series.extend_by_days(entry[1] - last_cumulative_time)
            last_cumulative_time = entry[1]
        last_timestep += 1
    if cumulative_time_error:
        log.error('cumulative times in Nexus summary file are not monotonically increasing')
    return time_series


def _open_file(summary_file):
    """Opens and reads the Nexus summary file into a list of entries.

    arguments:
        summary_file (str): path of Nexus output summary file

    returns:
        list of (timestep_number, time_in_days, date) where timestep_number is a non-negative int,
        time_in_days is a float, and date is a 10 charater date in format YYYY-MM-DD (or None if
        dates are not present in summary file)
    """

    us_date_format = True
    dates_present = True
    time_uom = 'd'
    summary_entries = []  # list of (time step no., cumulative time (days), date (dd/mm/yyyy))

    with open(summary_file, "r") as fp:
        while True:
            line = fp.readline()
            if len(line) == 0:
                break  # end of file
            words = line.split()
            if len(words) < 5:
                continue
            if not words[0].isdigit():
                if words[0].upper() == 'NO.':
                    assert words[3] in ['DAYS', 'HRS']
                    time_uom = 'd' if words[3] == 'DAYS' else 'h'
                    if len(words[4]) == 10 and words[4][2] == '/':
                        dates_present = True
                        date_format = words[4].upper()
                        us_date_format = (date_format == 'MM/DD/YYYY')
                        if not us_date_format:
                            assert date_format == 'DD/MM/YYYY', f'unrecognised date format in Nexus summary file: {date_format}'
                    else:
                        dates_present = False
                continue
            ts_number = int(words[0])
            time_in_days_or_hours = float(words[3])
            time_in_days = wam.convert(time_in_days_or_hours, time_uom, 'd', quantity = 'time')
            if dates_present:
                date_str = words[4]
                assert 9 <= len(date_str) <= 10, 'invalid date string length in summary file: ' + date_str
                if us_date_format:
                    date = dt.date(int(date_str[-4:]), int(date_str[:-8]), int(date_str[-7:-5]))
                else:
                    date = dt.date(int(date_str[-4:]), int(date_str[-7:-5]), int(date_str[:-8]))
            else:
                date = None
            summary_entries.append(
                (ts_number, time_in_days, date))  # TODO: check handling of dates and times, including sub-second times
    return summary_entries



"""Functions for finding grid cell faces to represent a surface."""

import logging

log = logging.getLogger(__name__)

import numpy as np
import warnings
import numba  # type: ignore
from numba import njit, prange  # type: ignore
from typing import Tuple, Optional, Dict

import resqpy as rq
import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.fault as rqf
import resqpy.property as rqp
import resqpy.weights_and_measures as wam
import resqpy.olio.box_utilities as bx
import resqpy.olio.intersection as meet
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec

# note: resqpy.grid_surface._grid_surface_cuda will be imported by the find_faces_to_represent_surface() function if needed


def find_faces_to_represent_surface_staffa(grid, surface, name, feature_type = "fault", progress_fn = None):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface.

    note:
        this version of the find faces function is designed for irregular grids
    """

    if progress_fn is not None:
        progress_fn(0.0)
    # log.debug('computing cell centres')
    centre_points = grid.centre_point()
    # log.debug('computing inter cell centre vectors and boxes')
    if grid.nk > 1:
        v = centre_points[:-1, :, :]
        u = centre_points[1:, :, :]
        k_vectors = u - v
        combo = np.stack((v, u))
        k_vector_boxes = np.empty((grid.nk - 1, grid.nj, grid.ni, 2, 3))
        k_vector_boxes[:, :, :, 0, :] = np.amin(combo, axis = 0)
        k_vector_boxes[:, :, :, 1, :] = np.amax(combo, axis = 0)
        column_k_vector_boxes = np.empty((grid.nj, grid.ni, 2, 3))
        column_k_vector_boxes[:, :, 0, :] = np.amin(k_vector_boxes[:, :, :, 0, :], axis = 0)
        column_k_vector_boxes[:, :, 1, :] = np.amax(k_vector_boxes[:, :, :, 1, :], axis = 0)
        k_faces = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
    else:
        k_vectors = None
        k_vector_boxes = None
        column_k_vector_boxes = None
        k_faces = None
    if grid.nj > 1:
        v = centre_points[:, :-1, :]
        u = centre_points[:, 1:, :]
        j_vectors = u - v
        combo = np.stack((v, u))
        j_vector_boxes = np.empty((grid.nk, grid.nj - 1, grid.ni, 2, 3))
        j_vector_boxes[:, :, :, 0, :] = np.amin(combo, axis = 0)
        j_vector_boxes[:, :, :, 1, :] = np.amax(combo, axis = 0)
        column_j_vector_boxes = np.empty((grid.nj - 1, grid.ni, 2, 3))
        column_j_vector_boxes[:, :, 0, :] = np.amin(j_vector_boxes[:, :, :, 0, :], axis = 0)
        column_j_vector_boxes[:, :, 1, :] = np.amax(j_vector_boxes[:, :, :, 1, :], axis = 0)
        j_faces = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
    else:
        j_vectors = None
        j_vector_boxes = None
        column_j_vector_boxes = None
        j_faces = None
    if grid.ni > 1:
        i_vectors = centre_points[:, :, 1:] - centre_points[:, :, :-1]
        v = centre_points[:, :, :-1]
        u = centre_points[:, :, 1:]
        i_vectors = u - v
        combo = np.stack((v, u))
        i_vector_boxes = np.empty((grid.nk, grid.nj, grid.ni - 1, 2, 3))
        i_vector_boxes[:, :, :, 0, :] = np.amin(combo, axis = 0)
        i_vector_boxes[:, :, :, 1, :] = np.amax(combo, axis = 0)
        column_i_vector_boxes = np.empty((grid.nj, grid.ni - 1, 2, 3))
        column_i_vector_boxes[:, :, 0, :] = np.amin(i_vector_boxes[:, :, :, 0, :], axis = 0)
        column_i_vector_boxes[:, :, 1, :] = np.amax(i_vector_boxes[:, :, :, 1, :], axis = 0)
        i_faces = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)
    else:
        i_vectors = None
        i_vector_boxes = None
        column_i_vector_boxes = None
        i_faces = None

    # log.debug('finding surface triangle boxes')
    t, p = surface.triangles_and_points()
    if not bu.matching_uuids(grid.crs_uuid, surface.crs_uuid):
        log.debug("converting from surface crs to grid crs")
        s_crs = rqc.Crs(surface.model, uuid = surface.crs_uuid)
        s_crs.convert_array_to(grid.crs, p)
    triangles = p[t]
    assert triangles.size > 0, "no triangles in surface"
    triangle_boxes = np.empty((triangles.shape[0], 2, 3))
    triangle_boxes[:, 0, :] = np.amin(triangles, axis = 1)
    triangle_boxes[:, 1, :] = np.amax(triangles, axis = 1)

    grid_box = grid.xyz_box(lazy = False, local = True)

    # log.debug('looking for cell faces for each triangle')
    batch_size = 1000
    triangle_count = triangles.shape[0]
    progress_batch = min(1.0, float(batch_size) / float(triangle_count))
    progress_base = 0.0
    ti_base = 0
    while ti_base < triangle_count:
        ti_end = min(ti_base + batch_size, triangle_count)
        batch_box = np.empty((2, 3))
        batch_box[0, :] = np.amin(triangle_boxes[ti_base:ti_end, 0, :], axis = 0)
        batch_box[1, :] = np.amax(triangle_boxes[ti_base:ti_end, 1, :], axis = 0)
        if bx.boxes_overlap(grid_box, batch_box):
            for j in range(grid.nj):
                if progress_fn is not None:
                    progress_fn(progress_base + progress_batch * (float(j) / float(grid.nj)))
                for i in range(grid.ni):
                    if column_k_vector_boxes is not None and bx.boxes_overlap(batch_box, column_k_vector_boxes[j, i]):
                        full_intersects = meet.line_set_triangles_intersects(
                            centre_points[:-1, j, i],
                            k_vectors[:, j, i],
                            triangles[ti_base:ti_end],
                            line_segment = True,
                        )
                        distilled_intersects, _, _ = meet.distilled_intersects(full_intersects)
                        k_faces[distilled_intersects, j, i] = True
                    if (j < grid.nj - 1 and column_j_vector_boxes is not None and
                            bx.boxes_overlap(batch_box, column_j_vector_boxes[j, i])):
                        full_intersects = meet.line_set_triangles_intersects(
                            centre_points[:, j, i],
                            j_vectors[:, j, i],
                            triangles[ti_base:ti_end],
                            line_segment = True,
                        )
                        distilled_intersects, _, _ = meet.distilled_intersects(full_intersects)
                        j_faces[distilled_intersects, j, i] = True
                    if (i < grid.ni - 1 and column_i_vector_boxes is not None and
                            bx.boxes_overlap(batch_box, column_i_vector_boxes[j, i])):
                        full_intersects = meet.line_set_triangles_intersects(
                            centre_points[:, j, i],
                            i_vectors[:, j, i],
                            triangles[ti_base:ti_end],
                            line_segment = True,
                        )
                        distilled_intersects, _, _ = meet.distilled_intersects(full_intersects)
                        i_faces[distilled_intersects, j, i] = True
        ti_base = ti_end
        # log.debug('triangles processed: ' + str(ti_base))
        # log.debug('interim face counts: K: ' + str(np.count_nonzero(k_faces)) +
        #                              '; J: ' + str(np.count_nonzero(j_faces)) +
        #                              '; I: ' + str(np.count_nonzero(i_faces)))
        progress_base = min(1.0, progress_base + progress_batch)

    # log.debug('face counts: K: ' + str(np.count_nonzero(k_faces)) +
    #                      '; J: ' + str(np.count_nonzero(j_faces)) +
    #                      '; I: ' + str(np.count_nonzero(i_faces)))
    gcs = rqf.GridConnectionSet(
        grid.model,
        grid = grid,
        k_faces = k_faces,
        j_faces = j_faces,
        i_faces = i_faces,
        feature_name = name,
        feature_type = feature_type,
        create_organizing_objects_where_needed = True,
    )

    if progress_fn is not None:
        progress_fn(1.0)

    return gcs


def find_faces_to_represent_surface_regular(
    grid,
    surface,
    name,
    title = None,
    centres = None,
    agitate = False,
    random_agitation = False,
    feature_type = "fault",
    progress_fn = None,
    consistent_side = False,
    return_properties = None,
):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface.

    arguments:
        grid (RegularGrid): the grid for which to create a grid connection set representation of the surface
        surface (Surface): the surface to be intersected with the grid
        name (str): the feature name to use in the grid connection set
        title (str, optional): the citation title to use for the grid connection set; defaults to name
        centres (numpy float array of shape (nk, nj, ni, 3), optional): precomputed cell centre points in
           local grid space, to avoid possible crs issues; required if grid's crs includes an origin (offset)?
        agitate (bool, default False): if True, the points of the surface are perturbed by a small
           offset, which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        random_agitation (bool, default False): if True, the agitation is by a small random distance; if False,
           a constant positive shift of 5.0e-6 is applied to x, y & z values; ignored if agitate is False
        feature_type (str, default 'fault'): 'fault', 'horizon' or 'geobody boundary'
        progress_fn (f(x: float), optional): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments
        consistent_side (bool, default False): if True, the cell pairs will be ordered so that all the first
           cells in each pair are on one side of the surface, and all the second cells on the other
        return_properties (list of str, optional): if present, a list of property arrays to calculate and
           return as a dictionary; recognised values in the list are 'offset' and 'normal vector'; offset
           is a measure of the distance between the centre of the cell face and the intersection point of the
           inter-cell centre vector with a triangle in the surface; normal vector is a unit vector normal
           to the surface triangle; each array has an entry for each face in the gcs; the returned dictionary
           has the passed strings as keys and numpy arrays as values

    returns:
        gcs  or  (gcs, dict)
        where gcs is a new GridConnectionSet with a single feature, not yet written to hdf5 nor xml created;
        dict is a dictionary mapping from property name to numpy array; 'offset' will map to a numpy float
        array of shape (gcs.count, ); 'normal vector' will map to a numpy float array of shape (gcs.count, 3)
        holding a unit vector normal to the surface for each of the faces in the gcs; the dict is only
        returned if a non-empty list has been passed as return_properties

    notes:
        use find_faces_to_represent_surface_regular_optimised() instead, for faster computation;
        this function can handle the surface and grid being in different coordinate reference systems, as
        long as the implicit parent crs is shared; no trimming of the surface is carried out here: for
        computational efficiency, it is recommended to trim first;
        organisational objects for the feature are created if needed;
        if grid has differing xy & z units, this is accounted for here when generating normal vectors, ie.
        true normal unit vectors are returned
    """

    assert isinstance(grid, grr.RegularGrid)
    assert grid.is_aligned
    return_normal_vectors = False
    return_offsets = False
    if return_properties:
        assert all([p in ["offset", "normal vector"] for p in return_properties])
        return_normal_vectors = "normal vector" in return_properties
        return_offsets = "offset" in return_properties

    if title is None:
        title = name

    if progress_fn is not None:
        progress_fn(0.0)

    log.debug(f"intersecting surface {surface.title} with regular grid {grid.title}")
    log.debug(f"grid extent kji: {grid.extent_kji}")

    grid_dxyz = (
        grid.block_dxyz_dkji[2, 0],
        grid.block_dxyz_dkji[1, 1],
        grid.block_dxyz_dkji[0, 2],
    )
    if centres is None:
        centres = grid.centre_point(use_origin = True)
    if consistent_side:
        log.debug("making all triangles clockwise")
        # note: following will shuffle order of vertices within t cached in surface
        surface.make_all_clockwise_xy(reorient = True)
    t, p = surface.triangles_and_points()
    assert t is not None and p is not None, f"surface {surface.title} is empty"
    if agitate:
        if random_agitation:
            p += 1.0e-5 * (np.random.random(p.shape) - 0.5)
        else:
            p += 5.0e-6
    log.debug(f"surface: {surface.title}; p0: {p[0]}; crs uuid: {surface.crs_uuid}")
    log.debug(f"surface min xyz: {np.min(p, axis = 0)}")
    log.debug(f"surface max xyz: {np.max(p, axis = 0)}")
    if not bu.matching_uuids(grid.crs_uuid, surface.crs_uuid):
        log.debug("converting from surface crs to grid crs")
        s_crs = rqc.Crs(surface.model, uuid = surface.crs_uuid)
        s_crs.convert_array_to(grid.crs, p)
        surface.crs_uuid = grid.crs.uuid
        log.debug(f"surface: {surface.title}; p0: {p[0]}; crs uuid: {surface.crs_uuid}")
        log.debug(f"surface min xyz: {np.min(p, axis = 0)}")
        log.debug(f"surface max xyz: {np.max(p, axis = 0)}")

    log.debug(f"centres min xyz: {np.min(centres.reshape((-1, 3)), axis = 0)}")
    log.debug(f"centres max xyz: {np.max(centres.reshape((-1, 3)), axis = 0)}")

    t_count = len(t)

    # todo: batch up either centres or triangles to reduce memory requirement for large models

    # K direction (xy projection)
    if grid.nk > 1:
        log.debug("searching for k faces")
        k_faces = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
        k_sides = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
        k_offsets = (np.full((grid.nk - 1, grid.nj, grid.ni), np.nan) if return_offsets else None)
        k_normals = (np.full((grid.nk - 1, grid.nj, grid.ni, 3), np.nan) if return_normal_vectors else None)
        k_centres = centres[0, :, :].reshape((-1, 3))
        k_hits = vec.points_in_triangles(p, t, k_centres, projection = "xy", edged = True).reshape(
            (t_count, grid.nj, grid.ni))
        del k_centres
        if consistent_side:
            cwt = vec.clockwise_triangles(p, t, projection = "xy") >= 0.0
        for k_t, k_j, k_i in np.stack(np.where(k_hits), axis = -1):
            xyz = meet.line_triangle_intersect(
                centres[0, k_j, k_i],
                centres[-1, k_j, k_i] - centres[0, k_j, k_i],
                p[t[k_t]],
                line_segment = True,
                t_tol = 1.0e-6,
            )
            if xyz is None:  # meeting point is outwith grid
                continue
            k_face = int((xyz[2] - centres[0, k_j, k_i, 2]) / grid_dxyz[2])
            if k_face == -1:  # handle rounding precision issues
                k_face = 0
            elif k_face == grid.nk - 1:
                k_face -= 1
            assert 0 <= k_face < grid.nk - 1
            k_faces[k_face, k_j, k_i] = True
            if consistent_side:
                k_sides[k_face, k_j, k_i] = cwt[k_t]
            if return_offsets:
                # compute offset as z diff between xyz and face
                k_offsets[k_face, k_j,
                          k_i] = xyz[2] - 0.5 * (centres[k_face, k_j, k_i, 2] + centres[k_face + 1, k_j, k_i, 2])
            if return_normal_vectors:
                k_normals[k_face, k_j, k_i] = vec.triangle_normal_vector(p[t[k_t]])
                # todo: if consistent side, could deliver information about horizon surface inversion
                if k_normals[k_face, k_j, k_i, 2] > 0.0:
                    k_normals[k_face, k_j, k_i] = -k_normals[k_face, k_j, k_i]  # -ve z hemisphere normal
        del k_hits
        log.debug(f"k face count: {np.count_nonzero(k_faces)}")
    else:
        k_faces = None
        k_sides = None
        k_normals = None

    if progress_fn is not None:
        progress_fn(0.3)

    # J direction (xz projection)
    if grid.nj > 1:
        log.debug("searching for j faces")
        j_faces = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
        j_sides = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
        j_offsets = (np.full((grid.nk, grid.nj - 1, grid.ni), np.nan) if return_offsets else None)
        j_normals = (np.full((grid.nk, grid.nj - 1, grid.ni, 3), np.nan) if return_normal_vectors else None)
        j_centres = centres[:, 0, :].reshape((-1, 3))
        j_hits = vec.points_in_triangles(p, t, j_centres, projection = "xz", edged = True).reshape(
            (t_count, grid.nk, grid.ni))
        del j_centres
        if consistent_side:
            cwt = vec.clockwise_triangles(p, t, projection = "xz") >= 0.0
        for j_t, j_k, j_i in np.stack(np.where(j_hits), axis = -1):
            xyz = meet.line_triangle_intersect(
                centres[j_k, 0, j_i],
                centres[j_k, -1, j_i] - centres[j_k, 0, j_i],
                p[t[j_t]],
                line_segment = True,
                t_tol = 1.0e-6,
            )
            if xyz is None:  # meeting point is outwith grid
                continue
            j_face = int((xyz[1] - centres[j_k, 0, j_i, 1]) / grid_dxyz[1])
            if j_face == -1:  # handle rounding precision issues
                j_face = 0
            elif j_face == grid.nj - 1:
                j_face -= 1
            assert 0 <= j_face < grid.nj - 1
            j_faces[j_k, j_face, j_i] = True
            if consistent_side:
                j_sides[j_k, j_face, j_i] = cwt[j_t]
            if return_offsets:
                # compute offset as y diff between xyz and face
                j_offsets[j_k, j_face,
                          j_i] = xyz[1] - 0.5 * (centres[j_k, j_face, j_i, 1] + centres[j_k, j_face + 1, j_i, 1])
            if return_normal_vectors:
                j_normals[j_k, j_face, j_i] = vec.triangle_normal_vector(p[t[j_t]])
                if j_normals[j_k, j_face, j_i, 2] > 0.0:
                    j_normals[j_k, j_face, j_i] = -j_normals[j_k, j_face, j_i]  # -ve z hemisphere normal
        del j_hits
        log.debug(f"j face count: {np.count_nonzero(j_faces)}")
    else:
        j_faces = None
        j_sides = None
        j_normals = None

    if progress_fn is not None:
        progress_fn(0.6)

    # I direction (yz projection)
    if grid.ni > 1:
        log.debug("searching for i faces")
        i_faces = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)
        i_sides = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)
        i_offsets = (np.full((grid.nk, grid.nj, grid.ni - 1), np.nan) if return_offsets else None)
        i_normals = (np.full((grid.nk, grid.nj, grid.ni - 1, 3), np.nan) if return_normal_vectors else None)
        i_centres = centres[:, :, 0].reshape((-1, 3))
        i_hits = vec.points_in_triangles(p, t, i_centres, projection = "yz", edged = True).reshape(
            (t_count, grid.nk, grid.nj))
        del i_centres
        if consistent_side:
            cwt = vec.clockwise_triangles(p, t, projection = "yz") >= 0.0
        for i_t, i_k, i_j in np.stack(np.where(i_hits), axis = -1):
            xyz = meet.line_triangle_intersect(
                centres[i_k, i_j, 0],
                centres[i_k, i_j, -1] - centres[i_k, i_j, 0],
                p[t[i_t]],
                line_segment = True,
                t_tol = 1.0e-6,
            )
            if xyz is None:  # meeting point is outwith grid
                continue
            i_face = int((xyz[0] - centres[i_k, i_j, 0, 0]) / grid_dxyz[0])
            if i_face == -1:  # handle rounding precision issues
                i_face = 0
            elif i_face == grid.ni - 1:
                i_face -= 1
            assert 0 <= i_face < grid.ni - 1
            i_faces[i_k, i_j, i_face] = True
            if consistent_side:
                i_sides[i_k, i_j, i_face] = cwt[i_t]
            if return_offsets:
                # compute offset as x diff between xyz and face
                i_offsets[i_k, i_j,
                          i_face] = xyz[0] - 0.5 * (centres[i_k, i_j, i_face, 0] + centres[i_k, i_j, i_face + 1, 0])
            if return_normal_vectors:
                i_normals[i_k, i_j, i_face] = vec.triangle_normal_vector(p[t[i_t]])
                if i_normals[i_k, i_j, i_face, 2] > 0.0:
                    i_normals[i_k, i_j, i_face] = -i_normals[i_k, i_j, i_face]  # -ve z hemisphere normal
        del i_hits
        log.debug(f"i face count: {np.count_nonzero(i_faces)}")
    else:
        i_faces = None
        i_sides = None
        i_normals = None

    if progress_fn is not None:
        progress_fn(0.9)

    if not consistent_side:
        k_sides = None
        j_sides = None
        i_sides = None

    log.debug("converting face sets into grid connection set")
    gcs = rqf.GridConnectionSet(
        grid.model,
        grid = grid,
        k_faces = k_faces,
        j_faces = j_faces,
        i_faces = i_faces,
        k_sides = k_sides,
        j_sides = j_sides,
        i_sides = i_sides,
        feature_name = name,
        feature_type = feature_type,
        title = title,
        create_organizing_objects_where_needed = True,
    )

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_offsets:
        k_offsets_list = (np.empty((0,)) if k_offsets is None else k_offsets[np.where(k_faces)])
        j_offsets_list = (np.empty((0,)) if j_offsets is None else j_offsets[np.where(j_faces)])
        i_offsets_list = (np.empty((0,)) if i_offsets is None else i_offsets[np.where(i_faces)])
        all_offsets = _all_offsets(grid.crs, k_offsets_list, j_offsets_list, i_offsets_list)
        log.debug(f"gcs count: {gcs.count}; all offsets shape: {all_offsets.shape}")
        assert all_offsets.shape == (gcs.count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_normal_vectors:
        k_normals_list = (np.empty((0, 3)) if k_normals is None else k_normals[np.where(k_faces)])
        j_normals_list = (np.empty((0, 3)) if j_normals is None else j_normals[np.where(j_faces)])
        i_normals_list = (np.empty((0, 3)) if i_normals is None else i_normals[np.where(i_faces)])
        all_normals = np.concatenate((k_normals_list, j_normals_list, i_normals_list), axis = 0)
        log.debug(f"gcs count: {gcs.count}; all normals shape: {all_normals.shape}")
        assert all_normals.shape == (gcs.count, 3)
        if grid.crs.xy_units != grid.crs.z_units:
            wam.convert_lengths(all_normals[:, 2], grid.crs.z_units, grid.crs.xy_units)
            all_normals = vec.unit_vectors(all_normals)

    if progress_fn is not None:
        progress_fn(1.0)

    # if returning properties, construct dictionary
    if return_properties:
        props_dict = {}
        if return_offsets:
            props_dict["offset"] = all_offsets
        if return_normal_vectors:
            props_dict["normal vector"] = all_normals
        return (gcs, props_dict)

    return gcs


def find_faces_to_represent_surface_regular_optimised(
    grid,
    surface,
    name,
    title = None,
    agitate = False,
    random_agitation = False,
    feature_type = "fault",
    is_curtain = False,
    progress_fn = None,
    return_properties = None,
    raw_bisector = False,
):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface.

    argumants:
        grid (RegularGrid): the grid for which to create a grid connection set representation of the surface;
           must be aligned, ie. I with +x, J with +y, K with +z and local origin of (0.0, 0.0, 0.0)
        surface (Surface): the surface to be intersected with the grid
        name (str): the feature name to use in the grid connection set
        title (str, optional): the citation title to use for the grid connection set; defaults to name
        agitate (bool, default False): if True, the points of the surface are perturbed by a small
           offset, which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        random_agitation (bool, default False): if True, the agitation is by a small random distance; if False,
           a constant positive shift of 5.0e-6 is applied to x, y & z values; ignored if agitate is False
        feature_type (str, default 'fault'): 'fault', 'horizon' or 'geobody boundary'
        is_curtain (bool, default False): if True, only the top layer of the grid is processed and the bisector
           property, if requested, is generated with indexable element columns
        progress_fn (f(x: float), optional): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments
        return_properties (List[str]): if present, a list of property arrays to calculate and
           return as a dictionary; recognised values in the list are 'triangle', 'depth', 'offset',
           'flange bool', 'grid bisector', or 'grid shadow';
           triangle is an index into the surface triangles of the triangle detected for the gcs face; depth is
           the z value of the intersection point of the inter-cell centre vector with a triangle in the surface;
           offset is a measure of the distance between the centre of the cell face and the intersection point;
           grid bisector is a grid cell boolean property holding True for the set of cells on one
           side of the surface, deemed to be shallower;
           grid shadow is a grid cell int8 property holding 0: cell neither above nor below a K face of the
           gridded surface, 1 cell is above K face(s), 2 cell is below K face(s), 3 cell is between K faces;
           the returned dictionary has the passed strings as keys and numpy arrays as values
        raw_bisector (bool, default False): if True and grid bisector is requested then it is left in a raw
           form without assessing which side is shallower (True values indicate same side as origin cell)

    returns:
        gcs  or  (gcs, gcs_props)
        where gcs is a new GridConnectionSet with a single feature, not yet written to hdf5 nor xml created;
        gcs_props is a dictionary mapping from requested return_properties string to numpy array

    notes:
        this function is designed for aligned regular grids only;
        this function can handle the surface and grid being in different coordinate reference systems, as
        long as the implicit parent crs is shared;
        no trimming of the surface is carried out here: for computational efficiency, it is recommended
        to trim first;
        organisational objects for the feature are created if needed;
        if the offset return property is requested, the implicit units will be the z units of the grid's crs
    """

    assert isinstance(grid, grr.RegularGrid)
    assert grid.is_aligned
    return_triangles = False
    return_depths = False
    return_offsets = False
    return_bisector = False
    return_shadow = False
    return_flange_bool = False
    if return_properties:
        assert all([
            p in [
                "triangle",
                "depth",
                "offset",
                "grid bisector",
                "grid shadow",
                "flange bool",
            ] for p in return_properties
        ])
        return_triangles = "triangle" in return_properties
        return_depths = "depth" in return_properties
        return_offsets = "offset" in return_properties
        return_bisector = "grid bisector" in return_properties
        return_shadow = "grid shadow" in return_properties
        return_flange_bool = "flange bool" in return_properties
        if return_flange_bool:
            return_triangles = True

    if title is None:
        title = name

    if progress_fn is not None:
        progress_fn(0.0)

    log.debug(f"intersecting surface {surface.title} with regular grid {grid.title}")
    # log.debug(f'grid extent kji: {grid.extent_kji}')

    grid_dxyz = (
        grid.block_dxyz_dkji[2, 0],
        grid.block_dxyz_dkji[1, 1],
        grid.block_dxyz_dkji[0, 2],
    )
    triangles, points = surface.triangles_and_points()
    assert (triangles is not None and points is not None), f"surface {surface.title} is empty"
    if agitate:
        if random_agitation:
            points += 1.0e-5 * (np.random.random(points.shape) - 0.5)
        else:
            points += 5.0e-6
    # log.debug(f'surface: {surface.title}; p0: {points[0]}; crs uuid: {surface.crs_uuid}')
    # log.debug(f'surface min xyz: {np.min(points, axis = 0)}')
    # log.debug(f'surface max xyz: {np.max(points, axis = 0)}')
    if not bu.matching_uuids(grid.crs_uuid, surface.crs_uuid):
        log.debug("converting from surface crs to grid crs")
        s_crs = rqc.Crs(surface.model, uuid = surface.crs_uuid)
        s_crs.convert_array_to(grid.crs, points)
        surface.crs_uuid = grid.crs.uuid
        # log.debug(f'surface: {surface.title}; p0: {points[0]}; crs uuid: {surface.crs_uuid}')
        # log.debug(f'surface min xyz: {np.min(points, axis = 0)}')
        # log.debug(f'surface max xyz: {np.max(points, axis = 0)}')

    nk = 1 if is_curtain else grid.nk
    # K direction (xy projection)
    if nk > 1:
        # log.debug("searching for k faces")
        k_faces = np.zeros((nk - 1, grid.nj, grid.ni), dtype = bool)
        k_triangles = np.full((nk - 1, grid.nj, grid.ni), -1, dtype = int)
        k_depths = np.full((nk - 1, grid.nj, grid.ni), np.nan)
        k_offsets = np.full((nk - 1, grid.nj, grid.ni), np.nan)
        p_xy = np.delete(points, 2, 1)

        k_hits = vec.points_in_triangles_aligned_optimised(grid.ni, grid.nj, grid_dxyz[0], grid_dxyz[1],
                                                           p_xy[triangles])

        axis = 2
        index1 = 1
        index2 = 2
        k_faces, k_offsets, k_triangles = intersect_numba(
            axis,
            index1,
            index2,
            k_hits,
            nk,
            points,
            triangles,
            grid_dxyz,
            k_faces,
            return_depths,
            k_depths,
            return_offsets,
            k_offsets,
            return_triangles,
            k_triangles,
        )
        del k_hits
        del p_xy
        log.debug(f"k face count: {np.count_nonzero(k_faces)}")
    else:
        k_faces = None
        k_triangles = None
        k_depths = None
        k_offsets = None

    if progress_fn is not None:
        progress_fn(0.3)

    # J direction (xz projection)
    if grid.nj > 1:
        # log.debug("searching for j faces")
        j_faces = np.zeros((nk, grid.nj - 1, grid.ni), dtype = bool)
        j_triangles = np.full((nk, grid.nj - 1, grid.ni), -1, dtype = int)
        j_depths = np.full((nk, grid.nj - 1, grid.ni), np.nan)
        j_offsets = np.full((nk, grid.nj - 1, grid.ni), np.nan)
        p_xz = np.delete(points, 1, 1)

        j_hits = vec.points_in_triangles_aligned_optimised(grid.ni, nk, grid_dxyz[0], grid_dxyz[2], p_xz[triangles])

        axis = 1
        index1 = 0
        index2 = 2
        j_faces, j_offsets, j_triangles = intersect_numba(
            axis,
            index1,
            index2,
            j_hits,
            grid.nj,
            points,
            triangles,
            grid_dxyz,
            j_faces,
            return_depths,
            j_depths,
            return_offsets,
            j_offsets,
            return_triangles,
            j_triangles,
        )
        del j_hits
        del p_xz
        if is_curtain and grid.nk > 1:  # expand arrays to all layers
            j_faces = np.repeat(j_faces, grid.nk, axis = 0)
            j_triangles = np.repeat(j_triangles, grid.nk, axis = 0)
            j_depths = np.repeat(j_depths, grid.nk, axis = 0)
            j_offsets = np.repeat(j_offsets, grid.nk, axis = 0)
        log.debug(f"j face count: {np.count_nonzero(j_faces)}")
    else:
        j_faces = None
        j_triangles = None
        j_depths = None
        j_offsets = None

    if progress_fn is not None:
        progress_fn(0.6)

    # I direction (yz projection)
    if grid.ni > 1:
        # log.debug("searching for i faces")
        i_faces = np.zeros((nk, grid.nj, grid.ni - 1), dtype = bool)
        i_triangles = np.full((nk, grid.nj, grid.ni - 1), -1, dtype = int)
        i_depths = np.full((nk, grid.nj, grid.ni - 1), np.nan)
        i_offsets = np.full((nk, grid.nj, grid.ni - 1), np.nan)
        p_yz = np.delete(points, 0, 1)

        i_hits = vec.points_in_triangles_aligned_optimised(grid.nj, nk, grid_dxyz[1], grid_dxyz[2], p_yz[triangles])

        axis = 0
        index1 = 0
        index2 = 1
        i_faces, i_offsets, i_triangles = intersect_numba(
            axis,
            index1,
            index2,
            i_hits,
            grid.ni,
            points,
            triangles,
            grid_dxyz,
            i_faces,
            return_depths,
            i_depths,
            return_offsets,
            i_offsets,
            return_triangles,
            i_triangles,
        )
        del i_hits
        del p_yz
        if is_curtain and grid.nk > 1:  # expand arrays to all layers
            # log.debug('expanding curtain faces')
            i_faces = np.repeat(i_faces, grid.nk, axis = 0)
            i_triangles = np.repeat(i_triangles, grid.nk, axis = 0)
            i_depths = np.repeat(i_depths, grid.nk, axis = 0)
            i_offsets = np.repeat(i_offsets, grid.nk, axis = 0)
        log.debug(f"i face count: {np.count_nonzero(i_faces)}")
    else:
        i_faces = None
        i_triangles = None
        i_depths = None
        i_offsets = None

    if progress_fn is not None:
        progress_fn(0.9)

    log.debug("converting face sets into grid connection set")
    gcs = rqf.GridConnectionSet(
        grid.model,
        grid = grid,
        k_faces = k_faces,
        j_faces = j_faces,
        i_faces = i_faces,
        k_sides = None,
        j_sides = None,
        i_sides = None,
        feature_name = name,
        feature_type = feature_type,
        title = title,
        create_organizing_objects_where_needed = True,
    )
    # log.debug('finished coversion to gcs')

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_triangles:
        # log.debug('preparing triangles array')
        k_tri_list = (np.empty((0,)) if k_triangles is None else k_triangles[_where_true(k_faces)])
        j_tri_list = (np.empty((0,)) if j_triangles is None else j_triangles[_where_true(j_faces)])
        i_tri_list = (np.empty((0,)) if i_triangles is None else i_triangles[_where_true(i_faces)])
        all_tris = np.concatenate((k_tri_list, j_tri_list, i_tri_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all triangles shape: {all_tris.shape}')
        assert all_tris.shape == (gcs.count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_depths:
        # log.debug('preparing depths array')
        k_depths_list = (np.empty((0,)) if k_depths is None else k_depths[_where_true(k_faces)])
        j_depths_list = (np.empty((0,)) if j_depths is None else j_depths[_where_true(j_faces)])
        i_depths_list = (np.empty((0,)) if i_depths is None else i_depths[_where_true(i_faces)])
        all_depths = np.concatenate((k_depths_list, j_depths_list, i_depths_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all depths shape: {all_depths.shape}')
        assert all_depths.shape == (gcs.count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_offsets:
        # log.debug('preparing offsets array')
        k_offsets_list = (np.empty((0,)) if k_offsets is None else k_offsets[_where_true(k_faces)])
        j_offsets_list = (np.empty((0,)) if j_offsets is None else j_offsets[_where_true(j_faces)])
        i_offsets_list = (np.empty((0,)) if i_offsets is None else i_offsets[_where_true(i_faces)])
        all_offsets = _all_offsets(grid.crs, k_offsets_list, j_offsets_list, i_offsets_list)
        # log.debug(f'gcs count: {gcs.count}; all offsets shape: {all_offsets.shape}')
        assert all_offsets.shape == (gcs.count,)

    if return_flange_bool:
        # log.debug('preparing flange array')
        flange_bool_uuid = surface.model.uuid(title = "flange bool",
                                              obj_type = "DiscreteProperty",
                                              related_uuid = surface.uuid)
        assert (flange_bool_uuid is not None), f"No flange bool property found for surface: {surface.title}"
        flange_bool = rqp.Property(surface.model, uuid = flange_bool_uuid)
        flange_array = flange_bool.array_ref()
        all_flange = np.take(flange_array, all_tris)
        assert all_flange.shape == (gcs.count,)

    # note: following is a grid cells property, not a gcs property
    if return_bisector:
        if is_curtain:
            log.debug("preparing columns bisector")
            bisector = column_bisector_from_faces((grid.nj, grid.ni), j_faces[0], i_faces[0])
            # log.debug('finished preparing columns bisector')
        else:
            log.debug("preparing cells bisector")
            bisector, is_curtain = bisector_from_faces(tuple(grid.extent_kji), k_faces, j_faces, i_faces, raw_bisector)
            if is_curtain:
                bisector = bisector[0]  # reduce to a columns property

    # note: following is a grid cells property, not a gcs property
    if return_shadow:
        log.debug("preparing cells shadow")
        shadow = shadow_from_faces(tuple(grid.extent_kji), k_faces)

    if progress_fn is not None:
        progress_fn(1.0)

    log.debug(f"finishing find_faces_to_represent_surface_regular_optimised for {name}")

    # if returning properties, construct dictionary
    if return_properties:
        props_dict = {}
        if return_triangles:
            props_dict["triangle"] = all_tris
        if return_depths:
            props_dict["depth"] = all_depths
        if return_offsets:
            props_dict["offset"] = all_offsets
        if return_bisector:
            props_dict["grid bisector"] = (bisector, is_curtain)
        if return_shadow:
            props_dict["grid shadow"] = shadow
        if return_flange_bool:
            props_dict["flange bool"] = all_flange
        return (gcs, props_dict)

    return gcs


def find_faces_to_represent_surface(grid, surface, name, mode = "auto", feature_type = "fault", progress_fn = None):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface.

    arguments:
        grid (RegularGrid): the grid for which to create a grid connection set representation of the surface
        surface (Surface): the triangulated surface for which grid cell faces are required
        name (str): the feature name to use in the grid connection set
        mode (str, default 'auto'): one of 'auto', 'staffa', 'regular', 'regular_optimised', 'regular_cuda';
           auto will translate to regular_optimised for regulat grids, and staffa for irregular grids;
           regular_cude required GPU hardware and the correct installation of numba.cuda and cupy
        feature_type (str, default 'fault'): 'fault', 'horizon' or 'geobody boundary'
        progress_fn (f(x: float), optional): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments

    returns:
        a new GridConnectionSet with a single feature, not yet written to hdf5 nor xml created

    note:
        this is a wrapper function selecting between more specialised functions; use those directly for more options
    """

    log.debug("finding cell faces for surface")
    if mode == "auto":
        if isinstance(grid, grr.RegularGrid) and grid.is_aligned:
            mode = "regular_optimised"
        else:
            mode = "staffa"
    if mode == "staffa":
        return find_faces_to_represent_surface_staffa(grid,
                                                      surface,
                                                      name,
                                                      feature_type = feature_type,
                                                      progress_fn = progress_fn)
    elif mode == "regular":
        return find_faces_to_represent_surface_regular(grid,
                                                       surface,
                                                       name,
                                                       feature_type = feature_type,
                                                       progress_fn = progress_fn)
    elif mode == "regular_optimised":
        return find_faces_to_represent_surface_regular_optimised(grid,
                                                                 surface,
                                                                 name,
                                                                 feature_type = feature_type,
                                                                 progress_fn = progress_fn)
    elif mode == "regular_cuda":
        import resqpy.grid_surface.grid_surface_cuda as rgs_c

        return rgs_c.find_faces_to_represent_surface_regular_cuda_mgpu(grid,
                                                                       surface,
                                                                       name,
                                                                       feature_type = feature_type,
                                                                       progress_fn = progress_fn)
    log.critical("unrecognised mode: " + str(mode))
    return None


def bisector_from_faces(  # type: ignore
    grid_extent_kji: Tuple[int, int, int],
    k_faces: np.ndarray,
    j_faces: np.ndarray,
    i_faces: np.ndarray,
    raw_bisector: bool,
) -> Tuple[np.ndarray, bool]:
    """Creates a boolean array denoting the bisection of the grid by the face sets.

    arguments:
        grid_extent_kji (Tuple[int, int, int]): the shape of the grid.
        k_faces (np.ndarray): a boolean array of which faces represent the surface in the k dimension.
        j_faces (np.ndarray): a boolean array of which faces represent the surface in the j dimension.
        i_faces (np.ndarray): a boolean array of which faces represent the surface in the i dimension.

    returns:
        Tuple containing:

        - array (np.ndarray): boolean bisectors array where values are True for cells on the side
            of the surface that has a lower mean k index on average and False for cells on the other side.
        - is_curtain (bool): True if the surface is a curtain (vertical), otherwise False.

    notes:
        The face sets must form a single 'sealed' cut of the grid (eg. not waving in and out of the grid).
        Any 'boxed in' parts of the grid (completely enclosed by bisecting faces) will be consistently
        assigned to either the True or False part.
    """
    assert len(grid_extent_kji) == 3

    # Finding the surface boundary (includes a buffer slice where surface does not reach edge of grid).
    boundary = get_boundary(k_faces, j_faces, i_faces, grid_extent_kji)

    # Setting up the bisector array for the bounding box.
    bounding_array = np.zeros(
        (
            boundary["k_max"] - boundary["k_min"] + 1,
            boundary["j_max"] - boundary["j_min"] + 1,
            boundary["i_max"] - boundary["i_min"] + 1,
        ),
        dtype = np.bool_,
    )

    # Seeding the bisector array from (0, 0, 0) up to the first faces that represent the surface.
    boundary_values = tuple(boundary.values())
    bounding_array, first_k, first_j, first_i = _seed_array((0, 0, 0), k_faces, j_faces, i_faces, boundary_values,
                                                            bounding_array)
    points = set()
    for dimension, first_true in enumerate([first_k, first_j, first_i]):
        for dimension_value in range(1, first_true):
            point = [0, 0, 0]
            point[dimension] = dimension_value
            point = tuple(point)  # type: ignore
            bounding_array, first_k_sub, first_j_sub, first_i_sub = _seed_array(point, k_faces, j_faces, i_faces,
                                                                                boundary_values, bounding_array)
            for sub_dimension, first_true_sub in enumerate([first_k_sub, first_j_sub, first_i_sub]):
                if dimension != sub_dimension:
                    for sub_dimension_value in range(1, first_true_sub):
                        point = [0, 0, 0]
                        point[dimension] = dimension_value
                        point[sub_dimension] = sub_dimension_value
                        point = tuple(point)  # type: ignore
                        if point not in points:
                            points.add(point)
                            bounding_array, _, _, _ = _seed_array(
                                point,
                                k_faces,
                                j_faces,
                                i_faces,
                                boundary_values,
                                bounding_array,
                            )

    # Setting up the array for the changing values.
    changing_array = np.zeros_like(bounding_array, dtype = np.bool_)

    # Repeatedly spreading True values to neighbouring cells that are not the other side of a face.
    open_k = np.logical_not(k_faces)[boundary["k_min"]:boundary["k_max"], boundary["j_min"]:boundary["j_max"] + 1,
                                     boundary["i_min"]:boundary["i_max"] + 1,]
    open_j = np.logical_not(j_faces)[boundary["k_min"]:boundary["k_max"] + 1, boundary["j_min"]:boundary["j_max"],
                                     boundary["i_min"]:boundary["i_max"] + 1,]
    open_i = np.logical_not(i_faces)[boundary["k_min"]:boundary["k_max"] + 1, boundary["j_min"]:boundary["j_max"] + 1,
                                     boundary["i_min"]:boundary["i_max"],]
    while True:
        changing_array[:] = False

        # k faces
        changing_array[1:, :, :] = np.logical_and(bounding_array[:-1, :, :], open_k)
        changing_array[:-1, :, :] = np.logical_or(changing_array[:-1, :, :],
                                                  np.logical_and(bounding_array[1:, :, :], open_k))

        # j faces
        changing_array[:, 1:, :] = np.logical_or(changing_array[:, 1:, :],
                                                 np.logical_and(bounding_array[:, :-1, :], open_j))
        changing_array[:, :-1, :] = np.logical_or(changing_array[:, :-1, :],
                                                  np.logical_and(bounding_array[:, 1:, :], open_j))

        # i faces
        changing_array[:, :, 1:] = np.logical_or(changing_array[:, :, 1:],
                                                 np.logical_and(bounding_array[:, :, :-1], open_i))
        changing_array[:, :, :-1] = np.logical_or(changing_array[:, :, :-1],
                                                  np.logical_and(bounding_array[:, :, 1:], open_i))

        changing_array[:] = np.logical_and(changing_array, np.logical_not(bounding_array))
        if np.count_nonzero(changing_array) == 0:
            break
        bounding_array = np.logical_or(bounding_array, changing_array)

    # Setting up the full bisectors array and assigning the bounding box values.
    array = np.zeros(grid_extent_kji, dtype = np.bool_)
    array[boundary["k_min"]:boundary["k_max"] + 1, boundary["j_min"]:boundary["j_max"] + 1,
          boundary["i_min"]:boundary["i_max"] + 1,] = bounding_array

    # Setting values outside of the bounding box.
    if boundary["k_max"] != grid_extent_kji[0] - 1 and np.any(bounding_array[-1, :, :]):
        array[boundary["k_max"] + 1:, :, :] = True
    if boundary["k_min"] != 0:
        array[:boundary["k_min"], :, :] = True
    if boundary["j_max"] != grid_extent_kji[1] - 1 and np.any(bounding_array[:, -1, :]):
        array[:, boundary["j_max"] + 1:, :] = True
    if boundary["j_min"] != 0:
        array[:, :boundary["j_min"], :] = True
    if boundary["i_max"] != grid_extent_kji[2] - 1 and np.any(bounding_array[:, :, -1]):
        array[:, :, boundary["i_max"] + 1:] = True
    if boundary["i_min"] != 0:
        array[:, :, :boundary["i_min"]] = True

    # Check all array elements are not the same.
    true_count = np.count_nonzero(array)
    cell_count = array.size
    assert (0 < true_count < cell_count), "Face set for surface is leaky or empty (surface does not intersect grid)."

    # Negate the array if it minimises the mean k and determine if the surface is a curtain.
    layer_cell_count = grid_extent_kji[1] * grid_extent_kji[2]
    array_k_sum = 0
    array_opposite_k_sum = 0
    is_curtain = False
    for k in range(grid_extent_kji[0]):
        array_layer_count = np.count_nonzero(array[k])
        array_k_sum += (k + 1) * array_layer_count
        array_opposite_k_sum += (k + 1) * (layer_cell_count - array_layer_count)
    array_mean_k = float(array_k_sum) / float(true_count)
    array_opposite_mean_k = float(array_opposite_k_sum) / float(cell_count - true_count)
    if array_mean_k > array_opposite_mean_k and not raw_bisector:
        array[:] = np.logical_not(array)
    if abs(array_mean_k - array_opposite_mean_k) <= 0.001:
        # log.warning('unable to determine which side of surface is shallower')
        is_curtain = True

    return array, is_curtain


def column_bisector_from_faces(grid_extent_ji: Tuple[int, int], j_faces: np.ndarray, i_faces: np.ndarray) -> np.ndarray:
    """Returns a numpy bool array denoting the bisection of the top layer of the grid by the curtain face sets.

    arguments:
        grid_extent_ji (pair of int): the shape of a layer of the grid
        j_faces, i_faces (numpy bool arrays): True where an internal grid face forms part of the
            bisecting surface, shaped for a single layer

    returns:
        numpy bool array of shape grid_extent_ji, set True for cells on one side of the face sets;
        set False for cells on othe side

    notes:
        the face sets must form a single 'sealed' cut of the grid (eg. not waving in and out of the grid);
        any 'boxed in' parts of the grid (completely enclosed by bisecting faces) will be consistently
        assigned to the False part;
        the resulting array is suitable for use as a grid property with indexable element of columns;
        the array is set True for the side of the curtain that contains cell [0, 0]
    """
    assert len(grid_extent_ji) == 2
    assert j_faces.ndim == 2 and i_faces.ndim == 2
    assert j_faces.shape == (grid_extent_ji[0] - 1, grid_extent_ji[1])
    assert i_faces.shape == (grid_extent_ji[0], grid_extent_ji[1] - 1)
    a = np.zeros(grid_extent_ji, dtype = np.bool_)  # initialise to False
    c = np.zeros(grid_extent_ji, dtype = np.bool_)  # cells changing
    open_j = np.logical_not(j_faces)
    open_i = np.logical_not(i_faces)
    # set one or more seeds; todo: more seeds to improve performance if needed
    a[0, 0] = True
    # repeatedly spread True values to neighbouring cells that are not the other side of a face
    # todo: check that following works when a dimension has extent 1
    limit = grid_extent_ji[0] * grid_extent_ji[1]
    for _ in range(limit):
        c[:] = False
        # j faces
        c[1:, :] = np.logical_or(c[1:, :], np.logical_and(a[:-1, :], open_j))
        c[:-1, :] = np.logical_or(c[:-1, :], np.logical_and(a[1:, :], open_j))
        # i faces
        c[:, 1:] = np.logical_or(c[:, 1:], np.logical_and(a[:, :-1], open_i))
        c[:, :-1] = np.logical_or(c[:, :-1], np.logical_and(a[:, 1:], open_i))
        c[:] = np.logical_and(c, np.logical_not(a))
        if np.count_nonzero(c) == 0:  # no more changes
            break
        a[:] = np.logical_or(a, c)
    if np.all(a):
        log.warning("curtain is leaky or misses grid when setting column bisector")
    # log.debug(f'returning bisector with count: {np.count_nonzero(a)} of {a.size}; shape: {a.shape}')
    return a


def shadow_from_faces(extent_kji, k_faces):
    """Returns a numpy int8 array indicating whether cells are above, below or between K faces.

    arguments:
        extent_kji (triple int): the shape of the grid
        k_faces (bool array): True where a K face is present; shaped (nk - 1, nj, ni)

    returns:
        numpy int8 array of shape extent_kji; values are: 0 neither above nor below a K face;
            1: above any K faces in the column; 2 below any K faces in the column;
            3: between K faces (one or more above and one or more below)
    """
    assert len(extent_kji) == 3
    limit = extent_kji[0] - 1  # maximum number of iterations needed to spead shadow
    shadow = np.zeros(extent_kji, dtype = np.int8)
    shadow[:-1] = np.where(k_faces, 1, 0)
    shadow[1:] += np.where(k_faces, 2, 0)
    for _ in range(limit):
        c = np.logical_and(shadow[:-1] == 0, shadow[1:] == 1)
        if np.count_nonzero(c) == 0:
            break
        shadow[:-1] = np.where(c, 1, shadow[:-1])
    for _ in range(limit):
        c = np.logical_and(shadow[1:] == 0, shadow[:-1] == 2)
        if np.count_nonzero(c) == 0:
            break
        shadow[1:] = np.where(c, 2, shadow[1:])
    for _ in range(limit):
        c = np.logical_and(shadow[:-1] >= 2, shadow[1:] == 1)
        if np.count_nonzero(c) == 0:
            break
        shadow[:-1] = np.where(c, 3, shadow[:-1])
        shadow[1:] = np.where(c, 3, shadow[1:])
    return shadow


def get_boundary(  # type: ignore
    k_faces: np.ndarray,
    j_faces: np.ndarray,
    i_faces: np.ndarray,
    grid_extent_kji: Tuple[int, int, int],
) -> Dict[str, int]:
    """Cretaes a dictionary of the indices that bound the surface (where the faces are True).

    arguments:
        k_faces (np.ndarray): a boolean array of which faces represent the surface in the k dimension
        j_faces (np.ndarray): a boolean array of which faces represent the surface in the j dimension
        i_faces (np.ndarray): a boolean array of which faces represent the surface in the i dimension
        grid_extent_kji (Tuple[int, int, int]): the shape of the grid

    returns:
        boundary (Dict[str, int]): a dictionary of the indices that bound the surface

    note:
        input faces arrays are for internal grid faces (ie. extent reduced by 1 in axis of faces);
        a buffer slice is included where the surface does not reach the edge of the grid
    """

    boundary = {
        "k_min": None,
        "k_max": None,
        "j_min": None,
        "j_max": None,
        "i_min": None,
        "i_max": None,
    }

    starting = True

    for f_i, faces in enumerate([k_faces, j_faces, i_faces]):

        # NB. k, j & i for rest of loop refer to indices of faces, regardless of which face set is being processed

        where_k, where_j, where_i = _where_true(faces)

        if len(where_k) == 0:
            continue

        min_k = where_k[0]  # optimisation if np.where() guaranteed to return results in natural order
        max_k = where_k[-1]
        # min_k = np.amin(where_k)
        # max_k = np.amax(where_k)
        min_j = np.amin(where_j)
        max_j = np.amax(where_j)
        min_i = np.amin(where_i)
        max_i = np.amax(where_i)

        # include cells on both sides of internal faces
        # and add buffer slice where edge of grid not reached by surface faces
        if f_i == 0:
            max_k += 1
        else:
            if min_k > 0:
                min_k -= 1
            if max_k < j_faces.shape[0] - 1:
                max_k += 1
        if f_i == 1:
            max_j += 1
        else:
            if min_j > 0:
                min_j -= 1
            if max_j < k_faces.shape[1] - 1:
                max_j += 1
        if f_i == 2:
            max_i += 1
        else:
            if min_i > 0:
                min_i -= 1
            if max_i < k_faces.shape[2] - 1:
                max_i += 1

        if starting:
            boundary["k_min"] = min_k
            boundary["k_max"] = max_k
            boundary["j_min"] = min_j
            boundary["j_max"] = max_j
            boundary["i_min"] = min_i
            boundary["i_max"] = max_i
            starting = False
        else:
            if min_k < boundary["k_min"]:
                boundary["k_min"] = min_k
            if max_k > boundary["k_max"]:
                boundary["k_max"] = max_k
            if min_j < boundary["j_min"]:
                boundary["j_min"] = min_j
            if max_j > boundary["j_max"]:
                boundary["j_max"] = max_j
            if min_i < boundary["i_min"]:
                boundary["i_min"] = min_i
            if max_i > boundary["i_max"]:
                boundary["i_max"] = max_i

    return boundary  # type: ignore


@njit  # pragma: no cover
def _where_true(data: np.ndarray):
    """Jitted NumPy 'where' function to improve performance on subsequent calls."""
    return np.where(data)


@njit  # pragma: no cover
def _first_true(array: np.ndarray) -> Optional[int]:  # type: ignore
    """Returns the index + 1 of the first True value in the array."""
    for idx, val in np.ndenumerate(array):
        if val:
            return idx[0] + 1
    return array.size


@njit  # pragma: no cover
def intersect_numba(
    axis: int,
    index1: int,
    index2: int,
    hits: np.ndarray,
    n_axis: int,
    points: np.ndarray,
    triangles: np.ndarray,
    grid_dxyz: Tuple[float],
    faces: np.ndarray,
    return_depths: bool,
    depths: np.ndarray,
    return_offsets: bool,
    offsets: np.ndarray,
    return_triangles: bool,
    triangle_per_face: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Finds the faces that intersect the surface in 3D.

    arguments:
        axis (int): axis number. Axis i is 0, j is 1, and k is 2.
        index1 (int): the first index. Axis i is 0, j is 0, and k is 1.
        index2 (int): the second index. Axis i is 1, j is 2, and k is 2.
        hits (np.ndarray): boolean array of grid centres that intersected the surface for the given
            axis.
        n_axis (int): number of cells in the axis.
        points (np.ndarray): array of all the surface node points in 3D.
        triangles (np.ndarray): array of all the points indices creating each triangle.
        grid_dxyz (Tuple[float]): tuple of a cell's thickness in each axis.
        faces (np.ndarray): boolean array of each cell face that can represent the surface.
        return_depths (bool): if True, an array of the depths is populated.
        depths (np.ndarray): array of the z values of the
            intersection point of the inter-cell centre vector with a triangle in the surface.
        return_offsets (bool): if True, an array of the offsets is calculated.
        offsets (np.ndarray): array of the distance between the centre of the cell face and the
            intersection point of the inter-cell centre vector with a triangle in the surface.
        return_triangles (bool): if True, an array of triangle indices is returned.

    returns:
        Tuple containing:

        - faces (np.ndarray): boolean array of each cell face that can represent the surface.
        - offsets (np.ndarray): array of the distance between the centre of the cell face and the
            intersection point of the inter-cell centre vector with a triangle in the surface.
        - triangle_per_face (np.ndarray): array of triangle numbers
    """
    n_faces = faces.shape[2 - axis]
    for i in prange(len(hits)):
        tri, d1, d2 = hits[i]

        # Line start point in 3D which had a projection hit.
        centre_point_start = np.zeros(3, dtype = np.float64) + grid_dxyz[axis] / 2
        centre_point_start[2 - index1] = (d1 + 0.5) * grid_dxyz[2 - index1]
        centre_point_start[2 - index2] = (d2 + 0.5) * grid_dxyz[2 - index2]

        # Line end point in 3D.
        centre_point_end = np.copy(centre_point_start)
        centre_point_end[axis] = grid_dxyz[axis] * (n_axis - 0.5)

        xyz = meet.line_triangle_intersect_numba(
            centre_point_start,
            centre_point_end - centre_point_start,
            points[triangles[tri]],
            line_segment = True,
            t_tol = 1.0e-6,
        )
        if xyz is None:  # meeting point is outwith grid
            continue

        # The face corresponding to the grid and surface intersection at this point.
        face = int((xyz[axis] - centre_point_start[axis]) / grid_dxyz[axis])
        if face == -1:  # handle rounding precision issues
            face = 0
        elif face == n_faces:
            face -= 1
        assert 0 <= face < n_faces

        face_idx = np.zeros(3, dtype = np.int32)
        face_idx[index1] = d1
        face_idx[index2] = d2
        face_idx[2 - axis] = face

        faces[face_idx[0], face_idx[1], face_idx[2]] = True

        if return_depths:
            depths[face_idx[0], face_idx[1], face_idx[2]] = xyz[2]
        if return_offsets:
            offsets[face_idx[0], face_idx[1], face_idx[2]] = xyz[axis] - ((face + 1) * grid_dxyz[axis])
        if return_triangles:
            triangle_per_face[face_idx[0], face_idx[1], face_idx[2]] = tri

    return faces, offsets, triangle_per_face


@njit  # pragma: no cover
def _seed_array(
    point: Tuple[int, int, int],
    k_faces: np.ndarray,
    j_faces: np.ndarray,
    i_faces: np.ndarray,
    boundary: Tuple[int, int, int, int, int, int],
    array: np.ndarray,
) -> Tuple[np.ndarray, int, int, int]:
    """Sets values of the array True up until a face is hit in each direction.

    arguments:
        point (Tuple[int, int, int]): coordinates of the initial seed point.
        k_faces (np.ndarray): a boolean array of which faces represent the surface in the k dimension.
        j_faces (np.ndarray): a boolean array of which faces represent the surface in the j dimension.
        i_faces (np.ndarray): a boolean array of which faces represent the surface in the i dimension.
        boundary (Tuple[int, int, int, int, int, int]): the boundaries of the surface given in the order
            (minimum k, maximum k, minimum j, maximum j, minimum i, maximum i).
        array (np.ndarray): boolean array that will be seeded.

    returns:
        Tuple containing:

        - array (np.ndarray): boolean array that has been seeded.
        - first_k (int): the index of the first k face in the k direction from the seed point or the
            array size in the k direction if there are no k faces.
        - first_j (int): the index of the first j face in the j direction from the seed point or the
            array size in the j direction if there are no j faces.
        - first_i (int): the index of the first i face in the i direction from the seed point or the
            array size in the i direction if there are no i faces.
    """
    k = point[0]
    j = point[1]
    i = point[2]

    first_k = 0
    if k == 0:
        first_k = _first_true(k_faces[boundary[0]:boundary[1], boundary[2] + j, boundary[4] + i])
        array[:first_k, j, i] = True

    first_j = 0
    if j == 0:
        first_j = _first_true(j_faces[boundary[0] + k, boundary[2]:boundary[3], boundary[4] + i])
        array[k, :first_j, i] = True

    first_i = 0
    if i == 0:
        first_i = _first_true(i_faces[boundary[0] + k, boundary[2] + j, boundary[4]:boundary[5]])
        array[k, j, :first_i] = True

    return array, first_k, first_j, first_i


def _all_offsets(crs, k_offsets_list, j_offsets_list, i_offsets_list):
    if crs.xy_units == crs.z_units:
        return np.concatenate((k_offsets_list, j_offsets_list, i_offsets_list), axis = 0)
    ji_offsets = np.concatenate((j_offsets_list, i_offsets_list), axis = 0)
    wam.convert_lengths(ji_offsets, crs.xy_units, crs.z_units)
    return np.concatenate((k_offsets_list, ji_offsets), axis = 0)



"""well_object_funcs.py: resqpy well module for functions that impact well objects"""

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import warnings
import os
import lasio
import numpy as np
import pandas as pd

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.property as rqp
import resqpy.weights_and_measures as bwam
import resqpy.well as rqw
import resqpy.olio.wellspec_keywords as wsk


def add_wells_from_ascii_file(model,
                              crs_uuid,
                              trajectory_file,
                              comment_character = '#',
                              space_separated_instead_of_csv = False,
                              well_col = 'WELL',
                              md_col = 'MD',
                              x_col = 'X',
                              y_col = 'Y',
                              z_col = 'Z',
                              length_uom = 'm',
                              md_domain = None,
                              drilled = False):
    """Creates new md datum, trajectory, interpretation and feature objects for each well in an ascii file.

    arguments:
       crs_uuid (uuid.UUID): the unique identifier of the coordinate reference system applicable to the x,y,z data;
          if None, a default crs will be created, making use of the length_uom and z_inc_down arguments
       trajectory_file (string): the path of the ascii file holding the well trajectory data to be loaded
       comment_character (string, default '#'): character deemed to introduce a comment in the trajectory file
       space_separated_instead_of_csv (boolean, default False): if True, the columns in the trajectory file are space
          separated; if False, comma separated
       well_col (string, default 'WELL'): the heading for the column containing well names
       md_col (string, default 'MD'): the heading for the column containing measured depths
       x_col (string, default 'X'): the heading for the column containing X (usually easting) data
       y_col (string, default 'Y'): the heading for the column containing Y (usually northing) data
       z_col (string, default 'Z'): the heading for the column containing Z (depth or elevation) data
       length_uom (string, default 'm'): the units of measure for the measured depths; should be 'm' or 'ft'
       md_domain (string, optional): the source of the original deviation data; may be 'logger' or 'driller'
       drilled (boolean, default False): True should be used for wells that have been drilled; False otherwise (planned,
          proposed, or a location being studied)
       z_inc_down (boolean, default True): indicates whether z values increase with depth; only used in the creation
          of a default coordinate reference system; ignored if crs_uuid is not None

    returns:
       tuple of lists of objects: (feature_list, interpretation_list, trajectory_list, md_datum_list)

    notes:
       ascii file must be table with first line being column headers, with columns for WELL, MD, X, Y & Z;
       actual column names can be set with optional arguments;
       all the objects are added to the model, with array data being written to the hdf5 file for the trajectories;
       the md_domain and drilled values are stored in the RESQML metadata but are only for human information and do not
       generally affect computations
    """

    assert md_col and x_col and y_col and z_col
    md_col = str(md_col)
    x_col = str(x_col)
    y_col = str(y_col)
    z_col = str(z_col)
    if crs_uuid is None:
        crs_uuid = model.crs_uuid
    assert crs_uuid is not None, 'coordinate reference system not found when trying to add wells'

    try:
        df = pd.read_csv(trajectory_file,
                         comment = comment_character,
                         delim_whitespace = space_separated_instead_of_csv)
        if df is None:
            raise Exception
    except Exception:
        log.error('failed to read ascii deviation survey file: ' + str(trajectory_file))
        raise
    if well_col and well_col not in df.columns:
        log.warning('well column ' + str(well_col) + ' not found in ascii trajectory file: ' + str(trajectory_file))
        well_col = None
    if well_col is None:
        for col in df.columns:
            if str(col).upper().startswith('WELL'):
                well_col = str(col)
                break
    else:
        well_col = str(well_col)
    assert well_col
    unique_wells = set(df[well_col])
    if len(unique_wells) == 0:
        log.warning('no well data found in ascii trajectory file: ' + str(trajectory_file))
        # note: empty lists will be returned, below

    feature_list = []
    interpretation_list = []
    trajectory_list = []
    md_datum_list = []

    for well_name in unique_wells:

        log.debug('importing well: ' + str(well_name))
        # create single well data frame (assumes measured depths increasing)
        well_df = df[df[well_col] == well_name]
        # create a measured depth datum for the well and add as part
        first_row = well_df.iloc[0]
        if first_row[md_col] == 0.0:
            md_datum = rqw.MdDatum(model,
                                   crs_uuid = crs_uuid,
                                   location = (first_row[x_col], first_row[y_col], first_row[z_col]))
        else:
            md_datum = rqw.MdDatum(model, crs_uuid = crs_uuid,
                                   location = (first_row[x_col], first_row[y_col], 0.0))  # sea level datum
        md_datum.create_xml(title = str(well_name))
        md_datum_list.append(md_datum)

        # create a well feature and add as part
        feature = rqo.WellboreFeature(model, feature_name = well_name)
        feature.create_xml()
        feature_list.append(feature)

        # create interpretation and add as part
        interpretation = rqo.WellboreInterpretation(model, is_drilled = drilled, wellbore_feature = feature)
        interpretation.create_xml(title_suffix = None)
        interpretation_list.append(interpretation)

        # create trajectory, write arrays to hdf5 and add as part
        trajectory = rqw.Trajectory(model,
                                    md_datum = md_datum,
                                    data_frame = well_df,
                                    length_uom = length_uom,
                                    md_domain = md_domain,
                                    represented_interp = interpretation,
                                    well_name = well_name)
        trajectory.write_hdf5()
        trajectory.create_xml(title = well_name)
        trajectory_list.append(trajectory)

    return (feature_list, interpretation_list, trajectory_list, md_datum_list)


def well_name(well_object, model = None):
    """Returns the 'best' citation title from the object or related well objects.

    arguments:
       well_object (object, uuid or root): Object for which a well name is required. Can be a
          Trajectory, WellboreInterpretation, WellboreFeature, BlockedWell, WellboreMarkerFrame,
          WellboreFrame, DeviationSurvey or MdDatum object
       model (model.Model, optional): required if passing a uuid or root; not recommended otherwise

    returns:
       string being the 'best' citation title to serve as a well name, form the object or some related objects

    note:
       xml and relationships must be established for this function to work
    """

    def better_root(model, root_a, root_b):
        a = rqet.citation_title_for_node(root_a)
        b = rqet.citation_title_for_node(root_b)
        if a is None or len(a) == 0:
            return root_b
        if b is None or len(b) == 0:
            return root_a
        a_digits = 0
        for c in a:
            if c.isdigit():
                a_digits += 1
        b_digits = 0
        for c in b:
            if c.isdigit():
                b_digits += 1
        if a_digits < b_digits:
            return root_b
        elif b_digits < a_digits:
            return root_a
        parts_like_a = model.parts(title = a)
        parts_like_b = model.parts(title = b)
        if len(parts_like_a) > len(parts_like_b):
            return root_b
        return root_a

    def best_root(model, roots_list):
        if len(roots_list) == 0:
            return None
        if len(roots_list) == 1:
            return roots_list[0]
        if len(roots_list) == 2:
            return better_root(model, roots_list[0], roots_list[1])
        return better_root(model, roots_list[0], best_root(model, roots_list[1:]))

    def best_root_for_object(well_object, model = None):

        if well_object is None:
            return None
        if model is None:
            model = well_object.model
        root_list = []
        obj_root = None
        obj_uuid = None
        obj_type = None
        traj_root = None

        if isinstance(well_object, str):
            obj_uuid = bu.uuid_from_string(well_object)
            assert obj_uuid is not None, 'well_name string argument could not be interpreted as uuid'
            well_object = obj_uuid
        if isinstance(well_object, bu.uuid.UUID):
            obj_uuid = well_object
            obj_root = model.root_for_uuid(obj_uuid)
            assert obj_root is not None, 'uuid not found in model when looking for well name'
            obj_type = rqet.node_type(obj_root)
        elif rqet.is_node(well_object):
            obj_root = well_object
            obj_type = rqet.node_type(obj_root)
            obj_uuid = rqet.uuid_for_part_root(obj_root)
        elif isinstance(well_object, rqw.Trajectory):
            obj_type = 'WellboreTrajectoryRepresentation'
            traj_root = well_object.root
        elif isinstance(well_object, rqo.WellboreFeature):
            obj_type = 'WellboreFeature'
        elif isinstance(well_object, rqo.WellboreInterpretation):
            obj_type = 'WellboreInterpretation'
        elif isinstance(well_object, rqw.BlockedWell):
            obj_type = 'BlockedWellboreRepresentation'
            if well_object.trajectory is not None:
                traj_root = well_object.trajectory.root
        elif isinstance(well_object, rqw.WellboreMarkerFrame):  # note: trajectory might be None
            obj_type = 'WellboreMarkerFrameRepresentation'
            if well_object.trajectory is not None:
                traj_root = well_object.trajectory.root
        elif isinstance(well_object, rqw.WellboreFrame):  # note: trajectory might be None
            obj_type = 'WellboreFrameRepresentation'
            if well_object.trajectory is not None:
                traj_root = well_object.trajectory.root
        elif isinstance(well_object, rqw.DeviationSurvey):
            obj_type = 'DeviationSurveyRepresentation'
        elif isinstance(well_object, rqw.MdDatum):
            obj_type = 'MdDatum'

        assert obj_type is not None, 'argument type not recognized for well_name'
        if obj_type.startswith('obj_'):
            obj_type = obj_type[4:]
        if obj_uuid is None:
            obj_uuid = well_object.uuid
            obj_root = model.root_for_uuid(obj_uuid)

        if obj_type == 'WellboreFeature':
            interp_parts = model.parts(obj_type = 'WellboreInterpretation')
            interp_parts = model.parts_list_filtered_by_related_uuid(interp_parts, obj_uuid)
            all_parts = interp_parts
            all_traj_parts = model.parts(obj_type = 'WellboreTrajectoryRepresentation')
            if interp_parts is not None:
                for part in interp_parts:
                    traj_parts = model.parts_list_filtered_by_related_uuid(all_traj_parts, model.uuid_for_part(part))
                    all_parts += traj_parts
            if all_parts is not None:
                root_list = [model.root_for_part(part) for part in all_parts]
        elif obj_type == 'WellboreInterpretation':
            feat_roots = model.roots(obj_type = 'WellboreFeature', related_uuid = obj_uuid)  # should return one root
            traj_roots = model.roots(obj_type = 'WellboreTrajectoryRepresentation', related_uuid = obj_uuid)
            root_list = feat_roots + traj_roots
        elif obj_type == 'WellboreTrajectoryRepresentation':
            interp_parts = model.parts(obj_type = 'WellboreInterpretation')
            interp_parts = model.parts_list_filtered_by_related_uuid(interp_parts, obj_uuid)
            all_parts = interp_parts
            all_feat_parts = model.parts(obj_type = 'WellboreFeature')
            if interp_parts is not None:
                for part in interp_parts:
                    feat_parts = model.parts_list_filtered_by_related_uuid(all_feat_parts, model.uuid_for_part(part))
                    all_parts += feat_parts
            if all_parts is not None:
                root_list = [model.root_for_part(part) for part in all_parts]
        elif obj_type in [
                'BlockedWellboreRepresentation', 'WellboreMarkerFrameRepresentation', 'WellboreFrameRepresentation'
        ]:
            if traj_root is None:
                traj_root = model.root(obj_type = 'WellboreTrajectoryRepresentation', related_uuid = obj_uuid)
            root_list = [best_root_for_object(traj_root, model = model)]
        elif obj_type == 'DeviationSurveyRepresentation':
            root_list = [best_root_for_object(model.root(obj_type = 'MdDatum', related_uuid = obj_uuid), model = model)]
        elif obj_type == 'MdDatum':
            pass

        root_list.append(obj_root)

        return best_root(model, root_list)

    title = rqet.citation_title_for_node(best_root_for_object(well_object, model = model))
    return 'WELL' if not title else title


def add_las_to_trajectory(las: lasio.LASFile, trajectory, realization = None, check_well_name = False):
    """Creates a WellLogCollection and WellboreFrame from a LAS file.

    arguments:
       las: an lasio.LASFile object
       trajectory: an instance of :class:`resqpy.well.Trajectory` .
       realization (integer): if present, the single realisation (within an ensemble)
          that this collection is for
       check_well_name (bool): if True, raise warning if LAS well name does not match
          existing wellborefeature citation title

    returns:
       collection, well_frame: instances of :class:`resqpy.property.WellLogCollection`
          and :class:`resqpy.well.WellboreFrame`

    note:
       in this current implementation, the first curve in the las object must be
       Measured Depths, not e.g. TVDSS
    """

    # Lookup relevant related resqml parts
    model = trajectory.model
    well_interp = trajectory.wellbore_interpretation
    well_title = well_interp.title

    if check_well_name and well_title != las.well.WELL.value:
        warnings.warn(f'LAS well title {las.well.WELL.value} does not match resqml tite {well_title}')

    # Create a new wellbore frame, using depth data from first curve in las file
    depth_values = np.array(las.index).copy()
    assert isinstance(depth_values, np.ndarray)
    las_depth_uom = bwam.rq_length_unit(las.curves[0].unit)

    # Ensure depth units are correct
    bwam.convert_lengths(depth_values, from_units = las_depth_uom, to_units = trajectory.md_uom)
    assert len(depth_values) > 0

    well_frame = rqw.WellboreFrame(
        parent_model = model,
        trajectory = trajectory,
        mds = depth_values,
        represented_interp = well_interp,
    )
    well_frame.write_hdf5()
    well_frame.create_xml()

    # Create a WellLogCollection in which to put logs
    collection = rqp.WellLogCollection(frame = well_frame, realization = realization)

    # Read in data from each curve in turn (skipping first curve which has depths)
    for curve in las.curves[1:]:

        collection.add_log(
            title = curve.mnemonic,
            data = curve.data,
            unit = curve.unit,
            realization = realization,
            write = False,
        )
        collection.write_hdf5_for_imported_list()
        collection.create_xml_for_imported_list_and_add_parts_to_model()

    return collection, well_frame


def add_blocked_wells_from_wellspec(model, grid, wellspec_file, usa_date_format = False):
    """Add a blocked well for each well in a Nexus WELLSPEC file.

    arguments:
       model (model.Model object): model to which blocked wells are added
       grid (grid.Grid object): grid against which wellspec data will be interpreted
       wellspec_file (string): path of ascii file holding Nexus WELLSPEC keyword and data
       usa_date_format (bool): mm/dd/yyyy (True) vs. dd/mm/yyyy (False)

    returns:
       int: count of number of blocked wells created

    notes:
       this function appends to the hdf5 file and creates xml for the blocked wells (but does not store epc);
       'simulation' trajectory and measured depth datum objects will also be created
    """

    well_list_dict = wsk.load_wellspecs(wellspec_file, column_list = None, usa_date_format = usa_date_format)

    count = 0
    for well in well_list_dict:
        log.info('processing well: ' + str(well))
        bw = rqw.BlockedWell(model,
                             grid = grid,
                             wellspec_file = wellspec_file,
                             well_name = well,
                             check_grid_name = True,
                             use_face_centres = True,
                             usa_date_format = usa_date_format)
        if not bw.node_count:  # failed to load from wellspec, eg. because of no perforations in grid
            log.warning('no wellspec data loaded for well: ' + str(well))
            continue
        bw.write_hdf5(model.h5_file_name(), mode = 'a', create_for_trajectory_if_needed = True)
        bw.create_xml(model.h5_uuid(), title = well)
        count += 1

    log.info(f'{count} blocked wells created based on wellspec file: {wellspec_file}')

    return count


def add_logs_from_cellio(blockedwell, cellio):
    """Creates a WellIntervalPropertyCollection for a given BlockedWell, using a given cell I/O file.

    arguments:
       blockedwell: a resqml blockedwell object
       cellio: an ascii file exported from RMS containing blocked well geometry and logs;
           must contain columns i_index, j_index and k_index, plus additional columns for logs to be imported
    """
    # Get the initial variables from the blocked well
    assert isinstance(blockedwell, rqw.BlockedWell), 'Not a blocked wellbore object'
    collection = rqp.WellIntervalPropertyCollection(frame = blockedwell)
    well_name = blockedwell.trajectory.title.split(" ")[0]
    grid = blockedwell.model.grid()

    # Read the cell I/O file to get the available columns (cols) and the data (data), and write into a dataframe
    with open(cellio, 'r') as fp:
        wellfound = False
        cols, data = [], []
        for line in fp.readlines():
            if line == "\n":
                wellfound = False  # Blankline signifies end of well data
            words = line.split()
            if wellfound:
                if len(words) > 2 and not words[0].isdigit():
                    cols.append(line)
                else:
                    if len(words) > 9:
                        assert len(cols) == len(words), 'Number of columns found should match header of file'
                        data.append(words)
            if len(words) == 3:
                if words[0].upper() == well_name.upper():
                    wellfound = True
        assert len(data) > 0 and len(cols) > 3, f"No data for well {well_name} found in file"
        df = pd.DataFrame(data = data, columns = [x.split()[0] for x in cols])
        df = df.apply(pd.to_numeric)
        # Get the cell_indices from the grid for the given i/j/k
        df['cell_indices'] = grid.natural_cell_indices(
            np.array((df['k_index'] - 1, df['j_index'] - 1, df['i_index'] - 1), dtype = int).T)
        df = df.drop(['i_index', 'j_index', 'k_index', 'x_in', 'y_in', 'z_in', 'x_out', 'y_out', 'z_out'], axis = 1)
    assert (df['cell_indices'] == blockedwell.cell_indices
           ).all(), 'Cell indices do not match between blocked well and log inputs'

    # Work out if the data columns are continuous, categorical or discrete
    type_dict = {}
    lookup_dict = {}
    for col in cols:
        words = col.split()
        if words[0] not in ['i_index', 'j_index', 'k_index', 'x_in', 'y_in', 'z_in', 'x_out', 'y_out', 'z_out']:
            if words[1] == 'unit1':
                type_dict[words[0]] = 'continuous'
            elif words[1] == 'DISC' and not words[0] == 'ZONES':
                type_dict[words[0]] = 'categorical'
                lookup_dict[words[0]] = lookup_from_cellio(col, blockedwell.model)
            elif words[1] == 'param' or words[0] == 'ZONES':
                type_dict[words[0]] = 'discrete'
            else:
                raise TypeError(f'unrecognised data type for {col}')

    # Loop over the columns, adding them to the blockedwell property collection
    for log in df.columns:
        if log not in ['cell_indices']:
            data_type = type_dict[log]
            if log == 'ZONES':
                data_type, dtype, null, discrete = 'discrete', int, -1, True
            elif data_type == 'continuous':
                dtype, null, discrete = float, np.nan, False
            else:
                dtype, null, discrete = int, -1, True
            if data_type == 'categorical':
                lookup_uuid = lookup_dict[log]  # For categorical data, find or generate a StringLookupTable
            else:
                lookup_uuid = None
            array_list = np.zeros((np.shape(blockedwell.grid_indices)), dtype = dtype)
            vals = list(df[log])
            for i, index in enumerate(blockedwell.cell_grid_link):
                if index == -1:
                    assert blockedwell.grid_indices[i] == -1
                    array_list[i] = null
                else:
                    if blockedwell.cell_indices[index] == list(df['cell_indices'])[index]:
                        array_list[i] = vals[index]
            collection.add_cached_array_to_imported_list(
                cached_array = array_list,
                source_info = '',
                keyword = f"{os.path.basename(cellio).split('.')[0]}.{blockedwell.trajectory.title}.{log}",
                discrete = discrete,
                uom = None,
                property_kind = None,
                facet = None,
                null_value = null,
                facet_type = None,
                realization = None)
            collection.write_hdf5_for_imported_list()
            collection.create_xml_for_imported_list_and_add_parts_to_model(string_lookup_uuid = lookup_uuid)


def lookup_from_cellio(line, model):
    """Create a StringLookup Object from a cell I/O row containing a categorical column name and details.

    arguments:
       line: a string from a cell I/O file, containing the column (log) name, type and categorical information
       model: the model to add the StringTableLookup to

    returns:
       uuid: the uuid of a StringTableLookup, either for a newly created table, or for an existing table if an identical one exists
    """
    lookup_dict = {}
    value, string = None, None
    # Generate a dictionary of values and strings
    for i, word in enumerate(line.split()):
        if i == 0:
            title = word
        elif not i < 2:
            if value is not None and string is not None:
                lookup_dict[value] = string
                value, string = None, None
            if value is None:
                value = int(word)
            else:
                if i == len(line.split()) - 1:
                    lookup_dict[value] = word
                else:
                    string = word

    # Check if a StringLookupTable already exists in the model, with the same name and values
    for existing_uuid in model.uuids(obj_type = 'StringTableLookup'):
        table = rqp.StringLookup(parent_model = model, uuid = existing_uuid)
        if table.title == title:
            if table.str_dict == lookup_dict:
                return table.uuid  # If the exact table exists, reuse it by returning the uuid

    # If no matching StringLookupTable exists, make a new one and return the uuid
    lookup = rqp.StringLookup(parent_model = model, int_to_str_dict = lookup_dict, title = title)
    lookup.create_xml(add_as_part = True)
    return lookup.uuid



"""Functions specific to Nexus units of measure."""

# Nexus is a trademark of Halliburton


def nexus_uom_for_quantity(nexus_unit_system, quantity, english_volume_flavour = None):
    """Returns RESQML uom string expected by Nexus for given quantity class and unit system.

    arguments:
        nexus_unit_system (str): one of 'METRIC', 'METKG/CM2', 'METBAR', 'LAB', or 'ENGLISH'
        quantity (str): the RESQML quantity class of interest; currently suppported:
            'length', 'area', 'volume', 'volume per volume', 'permeability rock',
            'time', 'thermodynamic temperature', 'mass per volume', 'pressure',
            'volume per time'
        english_volume_flavour (str, optional): only needed for ENGLISH unit system and volume,
            volume per volume, or volume per time quantity; one of 'PV', 'OVER PV', 'FVF', 'GOR',
            'surface gas rate', or 'saturation'; see notes regarding FVF, also regarding flow
            rates

    returns:
        str: the RESQML uom string for the units required by Nexus

    notes:
        transmissibility not yet catered for here, as RESQML has transmissibility units without a
        viscosity component;
        Nexus volume unit expectations vary depending on the data being handled, and sometimes also
        where in the Nexus input dataset the data is being entered;
        resqpy.weights_and_measures.valid_quantities() and valid_uoms() may also be of interest;
        in the ENHLISH unit system, Nexus expacts gas formation volume factors in bbl / 1000 ft3
        but that is not a valid RESQML uom – this function will return bbl/bbl for ENGLISH FVF;
        also be wary of pore volume units when using the medieval ENGLISH unit system: the OVER
        keyword expects different units than GRID or recurrent override input; ENGLISH fluid flow
        rates will be returned as bbl/d unless the flavour is specified as 'surface gas rate'
    """

    nexus_unit_system = nexus_unit_system.upper()
    assert nexus_unit_system in ['METRIC', 'METKG/CM2', 'METBAR', 'LAB', 'ENGLISH']
    # todo: add other quantities as needed
    assert quantity in [
        'length', 'area', 'volume', 'volume per volume', 'permeability rock', 'rock permeability', 'time',
        'thermodynamic temperature', 'mass per volume', 'pressure', 'volume per time'
    ]
    if quantity == 'rock permeability':
        quantity = 'permeability rock'

    if (nexus_unit_system == 'ENGLISH' and english_volume_flavour is not None):
        english_volume_flavour = english_volume_flavour.lower()
        if english_volume_flavour == 'pv':
            assert quantity == 'volume'
            return 'ft3'  # correct for main GRID array input or recurrent override
        elif english_volume_flavour == 'over pv':
            assert quantity == 'volume'
            return 'bbl'  # for static override of pv (Nexus OVER)
        elif english_volume_flavour == 'surface gas rate':
            assert quantity == 'volume per time'
            return '1000 ft3/d'
        assert quantity == 'volume per volume'
        if english_volume_flavour == 'fvf':
            return 'bbl/bbl'
        elif english_volume_flavour == 'gor':
            return '1000 ft3/bbl'
        else:
            assert english_volume_flavour == 'saturation'  # handled by default in dictionary

    d = {
        'METRIC': {
            'length': 'm',
            'area': 'm2',
            'volume': 'm3',
            'volume per volume': 'm3/m3',
            'permeability rock': 'mD',
            'time': 'd',
            'thermodynamic temperature': 'degC',
            'mass per volume': 'kg/m3',
            'pressure': 'kPa',
            'volume per time': 'm3/d'
        },
        'METKG/CM2': {
            'length': 'm',
            'area': 'm2',
            'volume': 'm3',
            'volume per volume': 'm3/m3',
            'permeability rock': 'mD',
            'time': 'd',
            'thermodynamic temperature': 'degC',
            'mass per volume': 'kg/m3',
            'pressure': 'kgf/cm2',
            'volume per time': 'm3/d'
        },
        'METBAR': {
            'length': 'm',
            'area': 'm2',
            'volume': 'm3',
            'volume per volume': 'm3/m3',
            'permeability rock': 'mD',
            'time': 'd',
            'thermodynamic temperature': 'degC',
            'mass per volume': 'kg/m3',
            'pressure': 'bar',
            'volume per time': 'm3/d'
        },
        'LAB': {
            'length': 'cm',
            'area': 'cm2',
            'volume': 'cm3',
            'volume per volume': 'cm3/cm3',
            'permeability rock': 'mD',
            'time': 'h',
            'thermodynamic temperature': 'degC',
            'mass per volume': 'g/cm3',
            'pressure': 'psi',
            'volume per time': 'cm3/h'
        },
        'ENGLISH': {
            'length': 'ft',
            'area': 'ft2',
            'volume': 'ft3',  # NB. Nexus expects bbl in some situations!
            'volume per volume': 'ft3/ft3',  # note: some special cases dealt with above
            'permeability rock': 'mD',
            'time': 'd',
            'thermodynamic temperature': 'degF',
            'mass per volume': 'lbm/ft3',
            'pressure': 'psi',
            'volume per time': 'bbl/d'  # surface gas is special case handled above
        }
    }

    return d[nexus_unit_system][quantity]



"""Time series classes and functions."""

__all__ = [
    'TimeSeries', 'GeologicTimeSeries', 'AnyTimeSeries', 'TimeDuration', 'selected_time_series', 'simplified_timestamp',
    'cleaned_timestamp', 'time_series_from_list', 'merge_timeseries_from_uuid', 'geologic_time_str',
    'timeframe_for_time_series_uuid', 'any_time_series', 'time_series_from_nexus_summary', 'check_timestamp',
    'colloquial_date'
]

from ._any_time_series import AnyTimeSeries
from ._geologic_time_series import GeologicTimeSeries
from ._time_duration import TimeDuration
from ._time_series import TimeSeries, check_timestamp, colloquial_date
from ._functions import selected_time_series, simplified_timestamp, cleaned_timestamp, time_series_from_list, \
    merge_timeseries_from_uuid, geologic_time_str, timeframe_for_time_series_uuid, any_time_series
from ._from_nexus_summary import time_series_from_nexus_summary

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Module holding functions relating to the mapping of welbore frame properties onto blocked wells."""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np

import resqpy.property as rqp
import resqpy.well as rqw


def blocked_well_frame_contributions_list(bw, wbf):
    """Returns wellbore frame contributions to each cell of a blocked well.

    arguments:
        bw (BlockedWell): the blocked well to map the wellbore frame onto
        wbf (WellboreFrame): the wellbore frame to map to the blocked well

    returns:
        list of list of (int, float, float) with one entry per blocked well cell, and each
            entry being a list of (wellbore frame interval index,
                                   fraction of wellbore frame interval in cell,
                                   fraction of cell's wellbore interval in wellbore frame interval)
    """

    wb_fraction_of_wbf = []
    for wb_ii in range(bw.node_count - 1):
        if bw.grid_indices[wb_ii] < 0:
            continue
        cell_entry_md = bw.node_mds[wb_ii]
        cell_exit_md = bw.node_mds[wb_ii + 1]
        cell_dmd = cell_exit_md - cell_entry_md  # wellbore interval length within cell
        entry_wbf_i, entry_wbf_fr = wbf.interval_for_md(bw.node_mds[wb_ii])
        exit_wbf_i, exit_wbf_fr = wbf.interval_for_md(bw.node_mds[wb_ii + 1])
        if (exit_wbf_i == -1 and exit_wbf_fr < 0.5) or (entry_wbf_i == -1 and entry_wbf_fr > 0.5):
            wb_fraction_of_wbf.append([])  # cell is above or below all frame intervals
            continue
        if entry_wbf_i == -1:
            assert entry_wbf_fr < 0.5
            entry_wbf_i = 0
            entry_wbf_fr = 0.0
        if exit_wbf_i == -1:
            assert exit_wbf_fr > 0.5
            exit_wbf_i = wbf.node_count - 2
            exit_wbf_fr = 1.0
        assert exit_wbf_i >= entry_wbf_i
        if exit_wbf_i == entry_wbf_i:  # single frame interval contributing to this cell
            assert exit_wbf_fr >= entry_wbf_fr
            if exit_wbf_fr == entry_wbf_fr:  # kissing point
                wb_fraction_of_wbf.append([])
                continue
            cell_fr = (min(cell_exit_md, wbf.node_mds[exit_wbf_i + 1]) -
                       max(cell_entry_md, wbf.node_mds[entry_wbf_i])) / cell_dmd
            assert 0.0 <= cell_fr <= 1.0 + 1.0e-6
            wb_fraction_of_wbf.append([(entry_wbf_i, exit_wbf_fr - entry_wbf_fr, cell_fr)])
            continue
        # more than one frame interval contributes to this cell
        cell_fr = (wbf.node_mds[entry_wbf_i + 1] - max(cell_entry_md, wbf.node_mds[entry_wbf_i])) / cell_dmd
        assert 0.0 <= cell_fr <= 1.0
        contrib_list = [(entry_wbf_i, 1.0 - entry_wbf_fr, cell_fr)]
        for wbf_i in range(entry_wbf_i + 1, exit_wbf_i):  # these frame intervals entirely within cell
            contrib_list.append((wbf_i, 1.0, (wbf.node_mds[wbf_i + 1] - wbf.node_mds[wbf_i]) / cell_dmd))
        cell_fr = (min(cell_exit_md, wbf.node_mds[exit_wbf_i + 1]) - wbf.node_mds[exit_wbf_i]) / cell_dmd
        assert 0.0 <= cell_fr <= 1.0
        contrib_list.append((exit_wbf_i, exit_wbf_fr, cell_fr))
        wb_fraction_of_wbf.append(contrib_list)

    return wb_fraction_of_wbf


def add_blocked_well_properties_from_wellbore_frame(bw,
                                                    frame_uuid = None,
                                                    property_kinds_list = None,
                                                    realization = None,
                                                    set_length = None,
                                                    set_perforation_fraction = None,
                                                    set_frame_interval = False):
    """Add properties to this blocked well by derivation from wellbore frame intervals properties.

    arguments:
        bw (BlockedWell): the blocked well to add properties to
        frame_uuid (UUID, optional): the uuid of the wellbore frame to source properties from; if None, a
            solitary wellbore frame relating to the same trajectory as the blocked well will be used
        property_kinds_list (list of str, optional): if present, a list of handled property kinds which
            are to be set from the wellbore frame properties; if None, any handled property kinds that
            are present for the wellbore frame will be used
        realization (int, optional): if present, wellbore frame properties will be filtered by this
            realization number and it will be assigned to the blocked well properties that are created
        set_length (bool, optional): if True, a length property will be generated based on active
            measured depth intervals; if None, will be set True if length is in list of property kinds
            being processed
        set_perforation_fraction (bool, optional): if True, a perforation fraction property will be
            created based on the fraction of the measured depth within a blocked well cell that is
            flagged as active, ie. perforated at some time; if None, it will be created only if length
            and permeability thickness are both absent
        set_frame_interval (bool, default False): if True, a static discrete property holding the index
            of the dominant active wellbore frame interval (per blocked well cell) is created

    returns:
        list of uuids of created property parts (does not include any copied time series object)

    notes:
        this method is designed to set up some blocked well properties based on similar properties
        already established on a special wellbore frame, mainly for perforations and open hole completions;
        frame_uuid should be specified if there are well logs in the dataset, or other wellbore frames;
        if a permeability thickness property is being set based on a wellbore frame property, the value
        is divided between blocked well cells based solely on measured depth interval lengths, without
        reference to grid properties such as net to gross ratio or permeability;
        titles will be the same as those used in the frame properties, and 'PPERF' for partial
        perforation;
        if set_frame_interval is True, the resulting property will be given a soft relationship with the
        wellbore frame (in addition to its supporting representation reference relationship with the
        blocked well); a null value of -1 is used where no active frame interval is present in a cell;
        units of measure will also be the same as those in the wellbore frame;
        this method only supports single grid blocked wells at present;
        blocked well and wellbore frame must be in the same model
    """

    # note: 'active' ia a static property kind used to indicate that a cell is perforated or otherwise
    # open to flow at some time, so needs to appear somewhere in well specification data for flow simulation;
    # 'well connection open' is a dynamic indicator used to identify time periods when a connection is open
    handled_property_kinds = [
        'active', 'well connection open', 'length', 'permeability length', 'skin', 'wellbore radius',
        'nonDarcy flow coefficient'
    ]

    def pk_pc_and_ts(pc, pk):
        """Returns a selective property collection based on property kind and a time series uuid if used."""
        pk_pc = rqp.selective_version_of_collection(pc, property_kind = pk)
        ts_uuids = pk_pc.time_series_uuid_list(sort_list = False)
        assert 0 <= len(ts_uuids) <= 1
        ts_uuid = None if len(ts_uuids) == 0 else ts_uuids[0]
        assert ts_uuid is not None or pk_pc.number_of_parts() == 1
        return pk_pc, ts_uuid

    assert len(bw.grid_list) == 1,  \
        'blocked well references more than one grid; not supported for frame properties'

    if frame_uuid is None:  # note: there could be multiple wellbore frames if well logs are in the dataset
        frame_uuid = bw.model.uuid(obj_type = 'WellboreFrameRepresentation', related_uuid = bw.trajectory.uuid)
    assert frame_uuid is not None, f'wellbore frame neither specified nor found for trajectory: {bw.trajectory.title}'

    if property_kinds_list is not None:
        for pk in property_kinds_list:
            assert pk in handled_property_kinds, f'blocked well from frame property kind not supported: {pk}'

    # establish a property collection for this blocked well
    wb_pc = bw.extract_property_collection()

    # load wellbore frame and its properties
    wbf = rqw.WellboreFrame(bw.model, uuid = frame_uuid)
    wbf_pc = wbf.extract_property_collection()
    wbf_pc = rqp.selective_version_of_collection(wbf_pc, indexable = 'intervals', realization = realization)
    if wbf_pc.number_of_parts() == 0:
        log.warning(f'no intervals properties found for wellbore frame: {wbf.title}; realization: {realization}')
        return []

    # determine the actual property kinds list that will be worked with
    pk_list = []
    wbf_pk_list = wbf_pc.property_kind_list(sort_list = False)
    for pk in wbf_pk_list:
        if ((property_kinds_list is None and pk in handled_property_kinds) or
            (property_kinds_list is not None and pk in property_kinds_list)):
            pk_list.append(pk)
    if len(pk_list) == 0:
        log.warning('no appropriate property kinds found for wellbore frame: {wbf.title}')

    if set_length is None:
        set_length = ('length' in pk_list)

    if set_perforation_fraction is None:
        set_perforation_fraction = (not set_length and 'permeability length' not in pk_list)

    log.debug(f'working list of property kinds for: {wbf.title}: {pk_list}')

    # find mapping of wellbpre frame onto blocked well cells, based on measured depths
    # list of entry per cell:
    # entry is list of triplets of (frame interval index, fraction of frame interval, fraction of cell md interval)
    wb_fraction_of_wbf = bw.frame_contributions_list(wbf)
    assert len(wb_fraction_of_wbf) == bw.cell_count

    # inherit static properties from wellbore frame to blocked well; resampling mehtod depends on property kind
    # TODO: make length and perforation fraction properties dynamic
    wb_active_array = None
    if 'active' in pk_list or set_length or set_perforation_fraction or set_frame_interval:
        wbf_p = wbf_pc.singleton(property_kind = 'active')
        assert wbf_p is not None, 'problem with active wellbore frame property'
        wbf_a = wbf_pc.cached_part_array_ref(wbf_p)
        wb_a = np.zeros(bw.cell_count, dtype = bool)
        length = np.zeros(bw.cell_count, dtype = float)
        pperf = np.zeros(bw.cell_count, dtype = float)
        dominant_wbf_interval = np.full(bw.cell_count, -1, dtype = int)
        ci = 0
        for wb_ii in range(bw.node_count - 1):
            if bw.grid_indices[wb_ii] < 0:
                continue
            i = ci
            ci += 1
            wbf_contributions = wb_fraction_of_wbf[i]
            max_active_cell_fr = 0.0
            max_active_wbf_ii = None
            for (wbf_ii, _, cell_fr) in wbf_contributions:
                if wbf_a[wbf_ii]:
                    wb_a[i] = True  # any contributing frame interval active makes cell active
                    pperf[i] += cell_fr
                    if cell_fr > max_active_cell_fr:
                        max_active_cell_fr = cell_fr
                        max_active_wbf_ii = wbf_ii
            if pperf[i] > 1.0:
                pperf[i] = 1.0
            if max_active_wbf_ii is not None:
                dominant_wbf_interval[i] = max_active_wbf_ii
            length[i] = pperf[i] * (bw.node_mds[wb_ii + 1] - bw.node_mds[wb_ii])
        wb_active_array = wb_a.copy()
        if set_frame_interval:
            # position this first to make it easy to get its uuid, for adding a soft relationship
            wb_pc.add_cached_array_to_imported_list(dominant_wbf_interval,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    'wellbore frame interval',
                                                    discrete = True,
                                                    property_kind = 'wellbore frame interval',
                                                    null_value = -1,
                                                    realization = realization,
                                                    indexable_element = 'cells')
        if 'active' in pk_list:
            wb_pc.add_cached_array_to_imported_list(wb_a,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    wbf_pc.citation_title_for_part(wbf_p),
                                                    discrete = True,
                                                    property_kind = 'active',
                                                    realization = realization,
                                                    indexable_element = 'cells')
        if set_length:
            md_uom = bw.trajectory.md_uom
            wb_pc.add_cached_array_to_imported_list(length,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    'LENGTH',
                                                    discrete = False,
                                                    uom = f'{md_uom}',
                                                    property_kind = 'length',
                                                    realization = realization,
                                                    indexable_element = 'cells')
        if set_perforation_fraction:
            md_uom = bw.trajectory.md_uom
            md_uom = 'ft' if md_uom.startswith('ft') else 'm'
            wb_pc.add_cached_array_to_imported_list(pperf,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    'PPERF',
                                                    discrete = False,
                                                    uom = f'{md_uom}/{md_uom}',
                                                    property_kind = 'perforation fraction',
                                                    realization = realization,
                                                    indexable_element = 'cells')

    if 'wellbore radius' in pk_list:
        wbf_p = wbf_pc.singleton(property_kind = 'wellbore radius')
        assert wbf_p is not None, 'problem with wellbore radius wellbore frame property'
        wbf_a = wbf_pc.cached_part_array_ref(wbf_p)
        wb_a = np.full(bw.cell_count, np.NaN, dtype = float)
        for i, wbf_contributions in enumerate(wb_fraction_of_wbf):
            if len(wbf_contributions) == 0:
                continue  # todo: could try to inherit wellbore radius from above or below?
            if np.all(np.isnan(wbf_a[wbf_contributions[0][0]:wbf_contributions[-1][0] + 1])):
                continue
            # use a maximum wellbore radius for the whole cell; over-optimistic in case of variation
            max_r = np.nanmax(wbf_a[wbf_contributions[0][0]:wbf_contributions[-1][0] + 1])
            if np.nanmin(wbf_a[wbf_contributions[0][0]:wbf_contributions[-1][0] + 1]) < max_r - 1.0e-6:
                log.warning(
                    f'radius of wellbore for {wbf.title} changes within cell {i} for blocked wellbore {bw.title}')
            # todo: find modal average or measured length weighted mean?; for now, use topmost (largest) radius
            wb_a[i] = max_r
        wb_pc.add_cached_array_to_imported_list(wb_a,
                                                f'derived from wellbore frame {wbf.title}',
                                                wbf_pc.citation_title_for_part(wbf_p),
                                                discrete = False,
                                                uom = wbf_pc.uom_for_part(wbf_p),
                                                property_kind = 'wellbore radius',
                                                realization = realization,
                                                indexable_element = 'cells')

    # if any dynamic, copy time series (possibly reusing an existing one, as likely to be shared with other wells)
    if 'well connection open' in pk_list:
        wco_pc, ts_uuid = pk_pc_and_ts(wbf_pc, 'well connection open')
        for wco_part in wco_pc.parts():
            wbf_wco = wco_pc.cached_part_array_ref(wco_part)
            assert wbf_wco is not None
            ti = wco_pc.time_index_for_part(wco_part)
            wb_wco = np.zeros(bw.cell_count, dtype = bool)
            for i, wbf_contributions in enumerate(wb_fraction_of_wbf):
                if len(wbf_contributions) == 0:
                    continue
                some_active_closed = False
                for (wbf_ii, _, _) in wbf_contributions:
                    if wbf_wco[wbf_ii]:
                        wb_wco[i] = True  # any contributing frame interval active makes cell active
                    elif wb_active_array is not None and wb_active_array[i]:
                        some_active_closed = True
                if wb_wco[i] and some_active_closed:
                    log.error('unsyncronised opening of active intervals ' +
                              f'at time index {ti} within cell {i} in blocked well {bw.title}')
            wb_pc.add_cached_array_to_imported_list(wb_wco,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    wco_pc.citation_title_for_part(wco_part),
                                                    discrete = True,
                                                    property_kind = 'well connection open',
                                                    time_index = ti,
                                                    time_series_uuid = ts_uuid,
                                                    realization = realization,
                                                    indexable_element = 'cells')

    if 'permeability length' in pk_list:
        kh_pc, ts_uuid = pk_pc_and_ts(wbf_pc, 'permeability length')
        kh_uom = None
        for kh_part in kh_pc.parts():
            wbf_kh = kh_pc.cached_part_array_ref(kh_part)
            assert wbf_kh is not None
            if kh_uom is None:
                kh_uom = kh_pc.uom_for_part(kh_part)
            else:
                assert kh_pc.uom_for_part(kh_part) == kh_uom, 'mixed permeability length units of measure'
            wb_kh = np.zeros(bw.cell_count, dtype = float)
            for i, wbf_contributions in enumerate(wb_fraction_of_wbf):
                if len(wbf_contributions) == 0:
                    continue
                for (wbf_ii, wbf_fr, _) in wbf_contributions:
                    if not np.isnan(wbf_kh[wbf_ii]):
                        wb_kh[i] += wbf_kh[wbf_ii] * wbf_fr
            wb_pc.add_cached_array_to_imported_list(wb_kh,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    kh_pc.citation_title_for_part(kh_part),
                                                    discrete = False,
                                                    uom = kh_uom,
                                                    property_kind = 'permeability length',
                                                    time_index = kh_pc.time_index_for_part(kh_part),
                                                    time_series_uuid = ts_uuid,
                                                    realization = realization,
                                                    indexable_element = 'cells')

    if 'skin' in pk_list:
        # note: here the skin contributions are weighted by fraction of measured length within a cell
        # this will not give a technically correct result if the permeability varies between frame intervals
        skin_pc, ts_uuid = pk_pc_and_ts(wbf_pc, 'skin')
        skin_uom = None  # probably always Euc
        for skin_part in skin_pc.parts():
            wbf_skin = skin_pc.cached_part_array_ref(skin_part)
            assert wbf_skin is not None
            if skin_uom is None:
                skin_uom = skin_pc.uom_for_part(skin_part)
            else:
                assert skin_pc.uom_for_part(skin_part) == skin_uom, 'mixed skin units of measure'
            wb_skin = np.full(bw.cell_count, np.NaN, dtype = float)
            for i, wbf_contributions in enumerate(wb_fraction_of_wbf):
                if len(wbf_contributions) == 0:
                    continue
                fr_sum = 0.0
                for (wbf_ii, _, cell_fr) in wbf_contributions:
                    if np.isnan(wbf_skin[wbf_ii]):
                        continue
                    if np.isnan(wb_skin[i]):
                        wb_skin[i] = 0.0
                    wb_skin[i] += wbf_skin[wbf_ii] * cell_fr
                    fr_sum += cell_fr
                if fr_sum > 0.0:
                    wb_skin[i] /= fr_sum
            wb_pc.add_cached_array_to_imported_list(wb_skin,
                                                    f'derived from wellbore frame {wbf.title}',
                                                    skin_pc.citation_title_for_part(skin_part),
                                                    discrete = False,
                                                    uom = skin_uom,
                                                    property_kind = 'skin',
                                                    time_index = skin_pc.time_index_for_part(skin_part),
                                                    time_series_uuid = ts_uuid,
                                                    realization = realization,
                                                    indexable_element = 'cells')

    # TODO: non Darcy flow coefficient
    # D factor is complicated: it appears in the inflow equation as a rate dependent adjustment to skin

    wb_pc.write_hdf5_for_imported_list()
    uuids = wb_pc.create_xml_for_imported_list_and_add_parts_to_model()

    if set_frame_interval:
        bw.model.create_reciprocal_relationship_uuids(uuids[0], 'sourceObject', frame_uuid, 'destinationObject')

    return uuids
